/* 
**    #   httpd.conf
**    LoadModule auth_mutual_module modules/mod_auth_mutual.so
**    <Location /auth_mutual>
**      AuthType mutual
**      AuthName "hello"
**      AuthUserFile /usr/local/apache-2.2/passwd/passwords
**      Require valid-user
**    </Location>
*/ 

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_log.h"
#include "ap_config.h"
#include "apr_strings.h"
#include "apr_lib.h"            /* for apr_isspace */
#include "apr_shm.h"
#include "apr_rmm.h"
#include "apr_proc_mutex.h"
#include "mod_auth.h"
#include "apr_strings.h"
#include "ap_provider.h"

#include "mutual.h"
#include "mod_auth_mutual.h"

#include "mod_ssl.h"

#if AP_MODULE_MAGIC_AT_LEAST(20100701,1)
#define HAS_ap_hook_note_auth_failure
#endif

module AP_MODULE_DECLARE_DATA auth_mutual_module;

#define AUTH_MUTUAL_VERSION "-draft07"

/*
 * Authentication protocol parameter names.
 */
#define AUTH_MUTUAL_SCHEME_NAME    "Mutual"
#define AUTH_MUTUAL_KEY_ALGORITHM  "algorithm"
#define AUTH_MUTUAL_KEY_VALIDATION "validation"
#define AUTH_MUTUAL_KEY_VERSION    "version"
#define AUTH_MUTUAL_KEY_USER       "user"
#define AUTH_MUTUAL_KEY_REALM      "realm"
#define AUTH_MUTUAL_KEY_PWD_HASH   "pwd-hash"
#define AUTH_MUTUAL_KEY_STALE      "stale"
#define AUTH_MUTUAL_KEY_SID        "sid"
#define AUTH_MUTUAL_KEY_NC         "nc"
#define AUTH_MUTUAL_KEY_NC_MAX     "nc-max"
#define AUTH_MUTUAL_KEY_NC_WND     "nc-window"
#define AUTH_MUTUAL_KEY_TIME       "time"
#define AUTH_MUTUAL_KEY_PATH       "path"
#define AUTH_MUTUAL_KEY_AUTH_DOMAIN "auth-domain"
#define AUTH_MUTUAL_KEY_WA         "wa"
#define AUTH_MUTUAL_KEY_WB         "wb"
#define AUTH_MUTUAL_KEY_OA         "oa"
#define AUTH_MUTUAL_KEY_OB         "ob"
#define AUTH_MUTUAL_KEY_LOCATION_WHEN_UNAUTHENTICATED "location-when-unauthenticated"
#define AUTH_MUTUAL_KEY_LOCATION_WHEN_LOGOUT          "location-when-logout"

/*
 * Request format classes
 */
enum req_class {
  REQ_A1,
  REQ_A3,
  REQ_ILLFORMED
};

/*
 * Response format classes
 */
enum res_class {
  RES_401_B0,
  RES_401_B0_STALE,
  RES_401_B1,
  RES_200_B4,
  RES_200_OPTIONAL
};

/* Request header.
 */
typedef struct header_in {
  enum req_class rclass;
  char *algorithm;
  const char *validation;
  char *version;
  char *auth_domain;
  char *user;
  char *wa;
  char *realm;
  char *sid;
  char *nc;
  auth_mutual_nc_t nc_value;
  int nc_is_valid;
  char *oa;
} header_in;

/* Response header
 */
typedef struct header_out {
  enum res_class rclass;
  int is_authenticated;
  const char *algorithm;
  const char *validation;
  const char *realm;
  char *pwd_hash;
  char *sid;
  char *wb;
  int         nc_max;
  int         nc_wnd;
  int         time;
  char *path;
  char *ob;
  char *auth_domain;
  char *login_url;
  char *logout_url;
} header_out;

enum validation_method {
  VALIDATION_HOST,
  VALIDATION_TLS_CERT,
  VALIDATION_TLS_KEY
};

static const char * const validation_strings[] = {
  "host", "tls-cert", "tls-key"
};

static void log_input_header(const request_rec *r, const header_in *p)
{
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                "Mutual: input header rclass=%d, algorithm=%s, validation=%s, user=%s, wa=%s, sid=%s, nc=%s, oa=%s",
                p->rclass, p->algorithm, p->validation, p->user, p->wa, p->sid, p->nc, p->oa);
}

static void log_mutual_command(const request_rec *r, const auth_mutual_command *p)
{
#ifdef APLOG_MODULE_IS_LEVEL
  int log = APLOG_MODULE_IS_LEVEL(r->server, APLOG_MODULE_INDEX, APLOG_DEBUG);
#else
  int log = r->server->loglevel >= APLOG_DEBUG;
#endif
  if (log) {
    char *s_data_in = NULL;
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Mutual: command algorithm=%d user=%s verifier=%s wa=%s oa=%s nc=%d validation=%s",
                  p->algorithm, p->user, p->verifier, p->wa, p->oa, p->nc, p->v);
    if (p->data_in) {
      s_data_in = auth_mutual_data_to_string(p->data_in);
      if (s_data_in) {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                      "Mutual:   with session[%s]", s_data_in);
        free(s_data_in);
      }
    }
  }
}

static void log_mutual_result(const request_rec *r, const auth_mutual_result *p)
{
#ifdef APLOG_MODULE_IS_LEVEL
  int log = APLOG_MODULE_IS_LEVEL(r->server, APLOG_MODULE_INDEX, APLOG_DEBUG);
#else
  int log = r->server->loglevel >= APLOG_DEBUG;
#endif
  if (log) {
#ifdef AUTH_MUTUAL_VERBOSE
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Mutual: result r=%d wb=%s ob=%s user=%s sb=%s",
                  p->result, p->wb, p->ob, p->user, p->sb);
#else
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Mutual: result r=%d wb=%s ob=%s user=%s",
                  p->result, p->wb, p->ob, p->user);
#endif
    if (p->data_out) {
      char *s_data_out = auth_mutual_data_to_string(p->data_out);
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "Mutual:   saving session[%s]", s_data_out);
      free(s_data_out);
    }
  }
}

/*
 * per Directory/Location configuration.
 */
typedef struct d_cfg {
  authn_provider_list *providers;
  char *dir;
  enum auth_mutual_algorithm algorithm;
  char *path;
  char *pwd_hash;
  char *optional_auth_user;
  char *auth_domain;
  int session_timeout;
  char *validation_host;
  char *login_url;
  char *logout_url;
} d_cfg;

static d_cfg *get_d_cfg(request_rec *r)
{
  return (d_cfg *)ap_get_module_config(r->per_dir_config, &auth_mutual_module);
}

/*
 * per request configuration.
 * saved in r->request_config
 *
 * Note: r_cfg is saved if request is the initial request or needs authentication.
 *       hdr_in  is NULL if request is subrequest.
 *       hdr_in  is not NULL if request is the initial request and has 'mutual' authentication header.
 *       hdr_out is not NULL if request needs authenticated by 'mutual' authentication method.
 */
typedef struct r_cfg {
  header_in  *hdr_in;
  header_out *hdr_out;
} r_cfg;

static r_cfg *get_r_cfg(request_rec *r)
{
  return (r_cfg *)ap_get_module_config(r->request_config, &auth_mutual_module);
}

static r_cfg *get_r_cfg_create(request_rec *r)
{
  r_cfg *rc = get_r_cfg(r);
  if (!rc) {
    rc = apr_pcalloc(r->pool, sizeof(*rc));
    ap_set_module_config(r->request_config, &auth_mutual_module, rc);
  }
  return rc;
}

static auth_mutual_state_t *get_m_state(server_rec *s)
{
  return (auth_mutual_state_t *)ap_get_module_config(s->module_config, &auth_mutual_module);
}

/*
 * import mod_ssl functions.
 */
static APR_OPTIONAL_FN_TYPE(ssl_is_https) *mutual_is_https = NULL;
static APR_OPTIONAL_FN_TYPE(ssl_var_lookup) *mutual_ssl_val = NULL;

static header_in *parse_header(request_rec *r, const char *auth_line);
static void verify_and_update_nonce(request_rec *r, header_in *hdr_in);

/*
 * Parse request header and check nonce.
 * This function hooks header_parser.
 *
 * We must verify nonce per request, not subrequest.
 * So, nonce checking should be done in header_parser hook.
 *
 * Parsed request header is stored in r_cfg.
 */
static int parse_header_and_check_nonce(request_rec *r)
{
  header_in *hdr_in;
  r_cfg *rc;
  const char *auth_line;

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: parse header hook: %s", r->uri);

  if (!ap_is_initial_req(r)) {
    /* this request is internal redirect */
    return DECLINED;
  }
  auth_line = apr_table_get(r->headers_in,
                            (PROXYREQ_PROXY == r->proxyreq)
                            ? "Proxy-Authorization" : "Authorization");
  if (!auth_line) {
    return DECLINED;
  }

  hdr_in = parse_header(r, auth_line);
  if (!hdr_in) {
    return DECLINED;
  }
  log_input_header(r, hdr_in);

  rc = get_r_cfg_create(r);
  rc->hdr_in  = hdr_in;
  rc->hdr_out = NULL;

  verify_and_update_nonce(r, hdr_in);

  return DECLINED;
}

/*
 * parse Authorization header strictly.
 * 
 */

static int is_token_char(unsigned char c) {
  if (c <= ' ' || '~' < c) return 0;
  switch (c) {
  case '(': case ')': case '<': case '>': case '@':
  case ',': case ';': case ':': case '\\': case '"':
  case '/': case '[': case ']': case '?': case '=':
  case '{': case '}':
    return 0;
  default:
    return 1;
  }
}

static char *next_token(apr_pool_t *p, const char **line)
{
  const char *s = *line;
  const char *t;

  while (*s == ' ' || *s == '\t') s++;
  t = s;
  while (is_token_char(*s)) s++;
  if (t < s) {
    *line = s;
    return apr_pstrndup(p, t, s - t);
  }
  return NULL;
}

static char *next_token_or_qstr(apr_pool_t *p, const char **line)
{
  const char *s = *line;

  while (*s == ' ' || *s == '\t') s++;
  if (*s != '"') {
    return next_token(p, line);
  } else {
    char *r;
    int i, j;
    int bs = 0, bsc = 0;

    for (i = 1; s[i] && (bs || s[i] != '"'); i++) {
      bs = !bs && s[i] == '\\';
      if (bs) bsc++;
    }
    if (!s[i]) return NULL;
    r = apr_pcalloc(p, i - bsc);
    for (i = 1, j = 0; s[i] && (bs || s[i] != '"'); i++) {
      bs = !bs && s[i] == '\\';
      if (!bs) r[j++] = s[i];
    }
    r[j] = 0;
    *line = s + i + 1;
    return r;
  }
}

/*
 * return 0 if SRC is not hex string.
 * return 1 if SRC is hex string
 *   *val = atoi(src)    if atoi(src) <= NC_MAX
 *   *val = NC_OVERFLOW  otherwise
 */
static int nc_parse(const char *src, auth_mutual_nc_t *val)
{
  int i;
  apr_int64_t nc;

  if (src[0] == '0' && src[1] != 0) return 0; /* leading zero is not allowed*/
  
  for (i = 0; src[i]; i++) {
    if (!apr_isxdigit(src[i]))
      return 0;
  }
  nc = apr_strtoi64(src, NULL, 10);
  if (nc < 0 || nc > AUTH_MUTUAL_NC_MAX) {
    *val = AUTH_MUTUAL_NC_OVERFLOW;
  } else {
    *val = (auth_mutual_nc_t)nc;
  }
  return 1;
}

static header_in *parse_header(request_rec *r, const char *auth_line)
{
  header_in *hdr_in;
  const char *p = auth_line;
  char *scheme;
  int nent;

  /*
   * RFC2617
   *
   * auth-line   = auth-scheme 1*SP 1#auth-param
   * auth-scheme = token
   * auth-param  = token "=" (token|quoted-string)
   */
  scheme = next_token(r->pool, &p);
  if (strcasecmp(scheme, AUTH_MUTUAL_SCHEME_NAME) != 0) {
    /* not auth_mutual auth header */
    return NULL;
  }
  
  hdr_in = apr_pcalloc(r->pool, sizeof(*hdr_in));
  hdr_in->rclass = REQ_ILLFORMED;
  hdr_in->algorithm  = NULL;
  hdr_in->validation = NULL;
  hdr_in->auth_domain = NULL;
  hdr_in->version = NULL;
  hdr_in->user = NULL;
  hdr_in->wa   = NULL;
  hdr_in->sid  = NULL;
  hdr_in->nc   = NULL;
  hdr_in->oa   = NULL;
  hdr_in->nc_value    = AUTH_MUTUAL_NC_OVERFLOW;
  hdr_in->nc_is_valid = 0;

  for (nent = 0; ; nent++) {
    char *key, *val;
    
    while (apr_isspace(*p) || *p == ',') p++;
    if (!*p) break;

    key = next_token(r->pool, &p);
    if (!key) return hdr_in;
    
    while (apr_isspace(*p)) p++;
    if (*p != '=') return hdr_in;
    p++;

    val = next_token_or_qstr(r->pool, &p);
    if (!val) return hdr_in;

    if (!strcasecmp(key, AUTH_MUTUAL_KEY_ALGORITHM) && !hdr_in->algorithm) {
      hdr_in->algorithm = val;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_VALIDATION) && !hdr_in->validation) {
      hdr_in->validation = val;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_AUTH_DOMAIN) && !hdr_in->auth_domain) {
      hdr_in->auth_domain = val;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_VERSION) && !hdr_in->version) {
      hdr_in->version = val;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_USER) && !hdr_in->user) {
      hdr_in->user = val;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_WA) && !hdr_in->wa) {
      hdr_in->wa = val;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_REALM) && !hdr_in->realm) {
      hdr_in->realm = val;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_SID) && !hdr_in->sid) {
      hdr_in->sid = val;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_NC) && !hdr_in->nc) {
      hdr_in->nc = val;
      if (!nc_parse(val, &hdr_in->nc_value))
        return hdr_in;
    } else if (!strcasecmp(key, AUTH_MUTUAL_KEY_OA) && !hdr_in->oa) {
      hdr_in->oa = val;
    } else {
      return hdr_in;
    }
  }
  if (hdr_in->version) nent--;
  if (hdr_in->auth_domain) nent--;
  if (hdr_in->algorithm && hdr_in->validation && hdr_in->user && hdr_in->wa && hdr_in->realm) {
    if (nent == 5)
      hdr_in->rclass = REQ_A1;
  }
  if (hdr_in->algorithm && hdr_in->validation && hdr_in->sid && hdr_in->realm && hdr_in->nc && hdr_in->oa) {
    if (nent == 6)
      hdr_in->rclass = REQ_A3;
  }
  return hdr_in;
}

static void verify_and_update_nonce(request_rec *r, header_in *hdr_in)
{
  if (hdr_in->rclass == REQ_A3) {
    auth_mutual_state_t *st = get_m_state(r->server);
    auth_mutual_session_state_t *p = auth_mutual_session_get(st, hdr_in->sid);
    if (p) {
      if (auth_mutual_validate_and_update_nonce(hdr_in->nc_value, &p->nc_rec)) {
        hdr_in->nc_is_valid = 1;
      } else {
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: invalid nonce sid=%s nc=%s, %s",
                      hdr_in->sid, hdr_in->nc, r->uri);
      }
      auth_mutual_session_release(st, p);
    } else {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: invalid session ID %s, %s", hdr_in->sid, r->uri);
    }
  }
}

static void note_auth_failure(request_rec *r, header_out *hdr_out);
static void save_auth_failure_string(request_rec *r, d_cfg *dc);
static header_out *auth_process(request_rec *r, header_in *hdr_in, d_cfg *dc);
static int well_configured(request_rec *r, d_cfg *dc);

/*
 * authenticate user.
 * This function hooks check_user_id.
 *
 */
static int auth_handler(request_rec *r)
{
  int ret;
  const char *t;
  request_rec *rmain;
  r_cfg *rc, *rcmain;
  header_in  *hdr_in;
  d_cfg *dc = get_d_cfg(r);
  
  t = ap_auth_type(r);
  if (!t || strcasecmp(t, AUTH_MUTUAL_SCHEME_NAME)) {
    return DECLINED;
  }

  r->ap_auth_type = AUTH_MUTUAL_SCHEME_NAME;

  if (!well_configured(r, dc)) {
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /*
   * get initial request
   */
  rmain = r;
  while (rmain->main) rmain = rmain->main;
  while (rmain->prev) rmain = rmain->prev;
  
  rcmain = get_r_cfg(rmain);
  hdr_in = NULL;
  if (rcmain) {
    hdr_in = rcmain->hdr_in;
  }

  rc = get_r_cfg_create(r);
  rc->hdr_out = auth_process(r, hdr_in, dc);

  if (!rc->hdr_out) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: auth_process failed %s", r->uri);
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  if (rc->hdr_out->is_authenticated) {
    save_auth_failure_string(r, dc);
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: auth success");
    return OK;
  } else {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: auth not complete %d: %s", rc->hdr_out->rclass, r->uri);
    note_auth_failure(r, rc->hdr_out);
    return HTTP_UNAUTHORIZED;
  }
}

static void default_header_out(request_rec *r, header_out *hdr_out, d_cfg *dc)
{
  hdr_out->rclass = RES_401_B0;
  hdr_out->is_authenticated = 0;
  hdr_out->algorithm = auth_mutual_algorithm_name(dc->algorithm);
  hdr_out->validation = NULL;
  hdr_out->realm = ap_auth_name(r);
  hdr_out->pwd_hash = dc->pwd_hash;
  hdr_out->sid = NULL;
  hdr_out->wb = NULL;
  hdr_out->nc_max = AUTH_MUTUAL_NC_MAX;
  hdr_out->nc_wnd = AUTH_MUTUAL_NC_WND;
  hdr_out->time = dc->session_timeout;
  hdr_out->path = dc->path ? dc->path : "/";
  hdr_out->ob = NULL;
  hdr_out->auth_domain = dc->auth_domain;
  hdr_out->login_url = dc->login_url;
  hdr_out->logout_url = dc->logout_url;
}

static int determine_validation(request_rec *r, d_cfg *dc,
                               enum validation_method *validation,
                               auth_mutual_command *cmd)
{
  int is_https = 0;

  if (mutual_is_https && mutual_is_https(r->connection)) {
    is_https = 1;
  }

  if (is_https) {
    char *p = NULL;
    if (mutual_ssl_val) {
      p = mutual_ssl_val(r->pool, r->server, r->connection, r, "SSL_SERVER_CERT");
    }
    if (!p) {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "Mutual: cannot get SSL_SERVER_CERT");
      return 0;
    }
    *validation = VALIDATION_TLS_CERT;
    if (cmd) {
      cmd->v = p;
      cmd->v_type = VALIDATION_KEY_PEM_CERT;
    }
    return 1;
  } else {
    *validation = VALIDATION_HOST;
    if (cmd) {
      cmd->v = dc->validation_host;
      cmd->v_type = VALIDATION_KEY_STRING;
    }
    return 1;
  }
}

static int is_quote_wb(const char *algorithm) {
  switch (auth_mutual_algorithm_value(algorithm)) {
  case AUTH_MUTUAL_ISO11770_4_DL_2048:
  case AUTH_MUTUAL_ISO11770_4_DL_4096:
    return 1;
  case AUTH_MUTUAL_ISO11770_4_EC_P256:
  case AUTH_MUTUAL_ISO11770_4_EC_P521:
    return 0;
  default:
    return 1;
  }
}

static char *auth_failure_message(request_rec *r, header_out *hdr_out, char **control)
{
  switch (hdr_out->rclass) {
  case RES_401_B0:
    if (control && hdr_out->login_url) {
      *control = apr_psprintf(r->pool,
                              "%s %s=\"%s\"",
                              AUTH_MUTUAL_SCHEME_NAME,
                              AUTH_MUTUAL_KEY_LOCATION_WHEN_UNAUTHENTICATED,
                              hdr_out->login_url);
    }
    /* FALLTHRU */
  case RES_401_B0_STALE:
  case RES_200_OPTIONAL:
    return apr_psprintf(r->pool,
                        "%s "         /* scheme      */
                        "%s=%s, "     /* algorithm   */
                        "%s=%s, "     /* validation  */
                        "%s=\"%s\", " /* realm       */
                        "%s=%s, "     /* version     */
                        "%s"          /* pwd-hash    */
                        "%s=%d"       /* stale       */
                        "%s",         /* auth-domain */
                        AUTH_MUTUAL_SCHEME_NAME,
                        AUTH_MUTUAL_KEY_ALGORITHM,  hdr_out->algorithm,
                        AUTH_MUTUAL_KEY_VALIDATION, hdr_out->validation,
                        AUTH_MUTUAL_KEY_REALM,      hdr_out->realm,
                        AUTH_MUTUAL_KEY_VERSION,    AUTH_MUTUAL_VERSION,
                        hdr_out->pwd_hash
                          ? apr_psprintf(r->pool, "%s=%s, ", AUTH_MUTUAL_KEY_PWD_HASH, hdr_out->pwd_hash)
                          : "",
                        AUTH_MUTUAL_KEY_STALE,
                        hdr_out->rclass == RES_401_B0_STALE? 1: 0,
                        hdr_out->auth_domain
                          ? apr_psprintf(r->pool, ", %s=\"%s\"", AUTH_MUTUAL_KEY_AUTH_DOMAIN, hdr_out->auth_domain)
                          : ""
                        );
  case RES_401_B1: {
    int quote_wb = is_quote_wb(hdr_out->algorithm);
    return apr_psprintf(r->pool,
                        "%s "         /* scheme      */
                        "%s=%s, "     /* algorithm   */
                        "%s=%s, "     /* validation  */
                        "%s=\"%s\", " /* realm       */
                        "%s=%s, "     /* version     */
                        "%s=%s, "     /* sid         */
                        "%s=%s%s%s, " /* wb          */
                        "%s=%d, "     /* nc-max      */
                        "%s=%d, "     /* nc-window   */
                        "%s=%d, "     /* time        */
                        "%s=\"%s\""   /* path        */
                        "%s",         /* auth-domain */
                        AUTH_MUTUAL_SCHEME_NAME,
                        AUTH_MUTUAL_KEY_ALGORITHM,  hdr_out->algorithm,
                        AUTH_MUTUAL_KEY_VALIDATION, hdr_out->validation,
                        AUTH_MUTUAL_KEY_REALM,  hdr_out->realm,
                        AUTH_MUTUAL_KEY_VERSION,    AUTH_MUTUAL_VERSION,
                        AUTH_MUTUAL_KEY_SID,    hdr_out->sid,
                        AUTH_MUTUAL_KEY_WB, (quote_wb ? "\"" : ""), hdr_out->wb, (quote_wb ? "\"" : ""),
                        AUTH_MUTUAL_KEY_NC_MAX, hdr_out->nc_max,
                        AUTH_MUTUAL_KEY_NC_WND, hdr_out->nc_wnd,
                        AUTH_MUTUAL_KEY_TIME,   hdr_out->time,
                        AUTH_MUTUAL_KEY_PATH,   hdr_out->path,
                        hdr_out->auth_domain
                          ? apr_psprintf(r->pool, ", %s=\"%s\"", AUTH_MUTUAL_KEY_AUTH_DOMAIN, hdr_out->auth_domain)
                          : ""
                        );
  }
  default:
    return NULL;
  }
}

static void note_auth_failure(request_rec *r, header_out *hdr_out)
{
  char *c = NULL;
  char *h = auth_failure_message(r, hdr_out, &c);

  apr_table_setn(r->err_headers_out,
                 (PROXYREQ_PROXY == r->proxyreq) ? "Proxy-Authenticate" : "WWW-Authenticate",
                 h);
  if (c) {
    apr_table_setn(r->err_headers_out, "Authentication-Control", c);
  }
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: response %s, %s", h, r->uri);
}

static void save_auth_failure_string(request_rec *r, d_cfg *dc)
{
#ifndef HAS_ap_hook_note_auth_failure
  header_out hdr_out;
  enum validation_method validation_method;
  char *control = NULL;
  apr_table_t *t = apr_table_make(r->pool, 2);
  char *h;
  
  default_header_out(r, &hdr_out, dc);
  if (!determine_validation(r, dc, &validation_method, NULL)) {
    return;
  }
  hdr_out.validation = validation_strings[validation_method];

  h = auth_failure_message(r, &hdr_out, &control);

  apr_table_setn(t,
                 (PROXYREQ_PROXY == r->proxyreq) ? "Proxy-Authenticate" : "WWW-Authenticate",
                 h);
  if (control) {
    apr_table_setn(t, "Authentication-Control", control);
  }
  apr_table_setn(r->notes, "note-auth-failure", (char *)t);
#endif
}

#ifdef HAS_ap_hook_note_auth_failure
static int hook_note_mutual_auth_failure(request_rec *r, const char *auth_type)
{
  d_cfg *dc;
  enum validation_method validation_method;
  header_out hdr_out;
  
  if (strcasecmp(auth_type, AUTH_MUTUAL_SCHEME_NAME)) {
    return DECLINED;
  }

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: hook_note_mutual_auth_failure");
  dc = get_d_cfg(r);
  default_header_out(r, &hdr_out, dc);
  if (!determine_validation(r, dc, &validation_method, NULL)) {
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  hdr_out.validation = validation_strings[validation_method];

  note_auth_failure(r, &hdr_out);
  return OK;
}
#endif

static authn_status get_hash(request_rec *r, const char *user, char **hash, d_cfg *dc)
{
  authn_status auth_result;
  authn_provider_list *current_provider;

  current_provider = dc->providers;
  do {
    const authn_provider *provider;
    
    /* For now, if a provider isn't set, we'll be nice and use the file
     * provider.
     */
    if (!current_provider) {
      provider = ap_lookup_provider(AUTHN_PROVIDER_GROUP,
                                    AUTHN_DEFAULT_PROVIDER, "0");
      
      if (!provider || !provider->get_realm_hash) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "Mutual: No Authn provider configured");
        auth_result = AUTH_GENERAL_ERROR;
        break;
      }
      apr_table_setn(r->notes, AUTHN_PROVIDER_NAME_NOTE, AUTHN_DEFAULT_PROVIDER);
    }
    else {
      provider = current_provider->provider;
      apr_table_setn(r->notes, AUTHN_PROVIDER_NAME_NOTE, current_provider->provider_name);
    }
    

    // TODO
    //   Since we record password in same format of Digest, we can use
    //   get_realm_hash as mod_auth_digest does. But, this format is temporary.
    
    /* We expect the password to be md5 hash of user:realm:password */
    auth_result = provider->get_realm_hash(r, user, ap_auth_name(r),
                                           hash);
    
    apr_table_unset(r->notes, AUTHN_PROVIDER_NAME_NOTE);

    /* Something occured.  Stop checking. */
    if (auth_result != AUTH_USER_NOT_FOUND) {
      break;
    }
    
    /* If we're not really configured for providers, stop now. */
    if (!dc->providers) {
      break;
    }

    current_provider = current_provider->next;
  } while (current_provider);
  
  return auth_result;
}

/*
 * process mutual authentication.
 *
 * hdr_in:  request header
 *
 * return response header or NULL if internal error.
 */
static header_out *auth_process(request_rec *r, header_in *hdr_in, d_cfg *dc)
{
  header_out *hdr_out;
  char *verifier = NULL;
  enum validation_method validation_method;
  int ret;
  auth_mutual_command cmd;
  auth_mutual_result result;
  auth_mutual_state_t *st = get_m_state(r->server);

  hdr_out = apr_pcalloc(r->pool, sizeof(*hdr_out));
  default_header_out(r, hdr_out, dc);
  
  if (!determine_validation(r, dc, &validation_method, &cmd)) {
    return NULL;
  }
  hdr_out->validation = validation_strings[validation_method];
  
  if (!hdr_in) {
    if (dc->optional_auth_user) {
      hdr_out->rclass = RES_200_OPTIONAL;
      hdr_out->is_authenticated = 1;
      r->user = apr_pstrdup(r->pool, dc->optional_auth_user);
      return hdr_out;
    } else {
      return hdr_out;
    }
  }

  if (hdr_in->rclass == REQ_ILLFORMED) {
    return hdr_out;
  }

  if (strcmp(hdr_in->algorithm,
             auth_mutual_algorithm_name(dc->algorithm)) != 0) {
    return hdr_out;
  }
  if (strcmp(hdr_in->realm, ap_auth_name(r)) != 0) {
    return hdr_out;
  }
  if (hdr_in->version && strcmp(hdr_in->version, AUTH_MUTUAL_VERSION) != 0) {
    return hdr_out;
  }
  if (hdr_out->auth_domain) {
    if (hdr_in->auth_domain == NULL || strcmp(hdr_in->auth_domain, hdr_out->auth_domain) != 0) {
      return hdr_out;
    }
  }
  
  switch (hdr_in->rclass) {
  case REQ_A1:
    cmd.req = AUTH_MUTUAL_REQ_A1;
    break;
  case REQ_A3:
    if (strcmp(hdr_in->validation, hdr_out->validation) != 0) {
      return hdr_out;
    }
    if (!hdr_in->nc_is_valid) {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "Mutual: nonce check failed %X: %s",
                    hdr_in->nc_value, r->uri);
      hdr_out->rclass = RES_401_B0_STALE;
      return hdr_out;
    }
    cmd.req = AUTH_MUTUAL_REQ_A3;
    break;
  }

  if (hdr_in->user != NULL) {
    authn_status authn_result = get_hash(r, hdr_in->user, &verifier, dc);
    if (authn_result != AUTH_USER_FOUND) {
      verifier = NULL;
    }
    if (authn_result == AUTH_USER_NOT_FOUND) {
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "Mutual: user `%s' in realm `%s' not found: %s",
                    hdr_in->user, ap_auth_name(r), r->uri);
    }
  }
  cmd.data_in = NULL;
  if (hdr_in->sid) {
    auth_mutual_session_state_t *p = auth_mutual_session_get(st, hdr_in->sid);
    if (p && strcmp(p->realm, ap_auth_name(r)) == 0) {
      cmd.data_in = auth_mutual_load_data(p->data, p->len);
      auth_mutual_session_release(st, p);
    }
  }
  cmd.algorithm = dc->algorithm;
  cmd.user = hdr_in->user;
  cmd.verifier = verifier;
  cmd.wa = hdr_in->wa;
  cmd.oa = hdr_in->oa;
  cmd.nc = hdr_in->nc_value;

  log_mutual_command(r, &cmd);
  ret = auth_mutual_process(&cmd, &result);
  
  auth_mutual_free_data(cmd.data_in);
  if (!ret) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "Mutual: auth_mutual_process failed.", r->uri);
    return NULL;
  }
  log_mutual_result(r, &result);

  if (result.invalidate_session && hdr_in->sid) {
    auth_mutual_session_state_t *p = auth_mutual_session_get(st, hdr_in->sid);
    if (p) {
      auth_mutual_invalidate_nonce(&p->nc_rec);
      auth_mutual_session_release(st, p);
    }
  }
  if (result.data_out) {
    size_t sz = auth_mutual_save_data(result.data_out, NULL, 0);
    if (sz > 0) {
      char hexsid[AUTH_MUTUAL_SID_HEX_LEN];
      
      auth_mutual_session_state_t *p = auth_mutual_session_alloc(st, dc->session_timeout, ap_auth_name(r), sz, hexsid);
      if (p) {
        auth_mutual_save_data(result.data_out, p->data, sz);
        hdr_out->sid = apr_pstrdup(r->pool, hexsid);
        auth_mutual_session_release(st, p);
      } else {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "Mutual: session memory exausted.", r->uri);
        auth_mutual_session_debug_dump(st, r);
        return NULL;
      }
    }
  }

  switch (result.result) {
  case AUTH_MUTUAL_RES_401_B0:
    hdr_out->rclass = RES_401_B0;
    break;
  case AUTH_MUTUAL_RES_401_B0_STALE:
    hdr_out->rclass = RES_401_B0_STALE;
    break;
  case AUTH_MUTUAL_RES_401_B1:
    hdr_out->rclass = RES_401_B1;
    hdr_out->wb = apr_pstrdup(r->pool, result.wb);
    break;
  case AUTH_MUTUAL_RES_200_B4:
    hdr_out->rclass = RES_200_B4;
    hdr_out->is_authenticated = 1;
    hdr_out->sid = hdr_in->sid;
    hdr_out->ob  = apr_pstrdup(r->pool, result.ob);
    r->user = apr_pstrdup(r->pool, result.user);
    break;
  default:
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  "Mutual: auth_mutual_process reutrns invalid result.", r->uri);
    auth_mutual_free_result(&result);
    return NULL;
  }
  auth_mutual_free_result(&result);
  return hdr_out;
}

static int well_configured(request_rec *r, d_cfg *dc)
{
  if (!ap_auth_name(r)) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "Mutual: need AuthName: %s", r->uri);
    return 0;
  }

  if (dc->validation_host == NULL) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "Mutual: 'host' validation .");
    return 0;
  }
  // TODO: check other settings.
  
  return 1;
}

static int is_quote_ob(const char *algorithm) {
  switch (auth_mutual_algorithm_value(algorithm)) {
  case AUTH_MUTUAL_ISO11770_4_DL_2048:
  case AUTH_MUTUAL_ISO11770_4_DL_4096:
    return 1;
  case AUTH_MUTUAL_ISO11770_4_EC_P256:
  case AUTH_MUTUAL_ISO11770_4_EC_P521:
    return 0;
  default:
    return 1;
  }
}

static int add_auth_info(request_rec *r)
{
  r_cfg *rc = get_r_cfg(r);

  if (rc && rc->hdr_out) {
    header_out *hdr_out = rc->hdr_out;
    char *key = NULL;
    char *h = NULL;
    if (hdr_out->rclass == RES_200_B4 && hdr_out->is_authenticated) {
      int quote_ob = is_quote_ob(hdr_out->algorithm);
      key = (PROXYREQ_PROXY == r->proxyreq) ? "Proxy-Authentication-Info" : "Authentication-Info";
      h = apr_psprintf(r->pool,
                       "%s "        /* scheme */
                       "%s=%s, "    /* sid    */
                       "%s=%s%s%s", /* ob     */
                       AUTH_MUTUAL_SCHEME_NAME,
                       AUTH_MUTUAL_KEY_SID, hdr_out->sid,
                       AUTH_MUTUAL_KEY_OB, (quote_ob ? "\"" : ""), hdr_out->ob, (quote_ob ? "\"" : ""));
      if (hdr_out->logout_url) {
        char *control;
        control = apr_psprintf(r->pool,
                               "%s %s=\"%s\"",
                               AUTH_MUTUAL_SCHEME_NAME,
                               AUTH_MUTUAL_KEY_LOCATION_WHEN_LOGOUT, hdr_out->logout_url);
        apr_table_setn(r->headers_out, "Authentication-Control", control);
      }
    } else if (rc->hdr_out->rclass == RES_200_OPTIONAL) {
      key = (PROXYREQ_PROXY == r->proxyreq) ? NULL : "Optional-WWW-Authenticate";
      h = auth_failure_message(r, hdr_out, NULL);
    }
    if (key && h) {
      apr_table_setn(r->headers_out, key, h);
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Mutual: response %s: %s, %s", key, h, r->uri);
    }
  }
  return OK;
}

static void *auth_create_config(apr_pool_t *p, server_rec *s)
{
  auth_mutual_state_t *st = (auth_mutual_state_t *)apr_pcalloc(p, sizeof(auth_mutual_state_t));
  st->cache_bytes = 1 * 1024 * 1024;
  st->cache_file = NULL;
  st->mutex_name = NULL;
  ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "Mutual: auth_create_config");

  auth_mutual_init_crypto_thread(p, s);
  auth_mutual_init();
  return st;
}

static int auth_post_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
  void *data;
  char *key = "auth_mutual_init";
  auth_mutual_state_t *st = get_m_state(s);

  mutual_is_https = APR_RETRIEVE_OPTIONAL_FN(ssl_is_https);
  mutual_ssl_val  = APR_RETRIEVE_OPTIONAL_FN(ssl_var_lookup);
  
  ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "Mutual: auth_post_config");
  /* auth_mutual_post_config() will be called twice. Don't bother
   * going through all of the initialization on the first call
   * because it will just be thrown away.*/
  apr_pool_userdata_get(&data, key, s->process->pool);
  if (!data) {
    apr_pool_userdata_set((void *)1, key, apr_pool_cleanup_null, s->process->pool);
    return OK;
  }

  return auth_mutual_session_init(st, p);
}

static void initialize_child(apr_pool_t *p, server_rec *s)
{
  ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "Mutual: auth_initialize_child");
  auth_mutual_state_t *st = get_m_state(s);
  auth_mutual_session_init_child(st, p);
}

static void auth_register_hooks(apr_pool_t *p)
{
  ap_hook_post_config(auth_post_config, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_child_init(initialize_child, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_header_parser(parse_header_and_check_nonce, NULL, NULL, APR_HOOK_MIDDLE);
  ap_hook_check_user_id(auth_handler, NULL, NULL, APR_HOOK_MIDDLE);
#ifdef HAS_ap_hook_note_auth_failure
  ap_hook_note_auth_failure(hook_note_mutual_auth_failure, NULL, NULL, APR_HOOK_MIDDLE);
#endif
  /*
   * XXX: add_auth_info adds Authentication-Info header.
   *
   * If some request 'r' requires authentication and  sub request 'rr' also requires it,
   * 'r' and 'rr' have 'Authentication-Info' header.
   * 'ap_internal_fast_redirect' doesn't merge this header. (mod_dir.c)
   * This causes duplicated 'Authentication-Info' header.
   *
   * To avoid this problem, we add the header at the last of fixups.
   */
  ap_hook_fixups(add_auth_info, NULL, NULL, APR_HOOK_REALLY_LAST);
}

static void *create_auth_mutual_dir_config(apr_pool_t *p, char *d)
{
  d_cfg *dc = apr_pcalloc(p, sizeof(*dc));

  dc->dir = d;
  dc->algorithm = AUTH_MUTUAL_ISO11770_4_EC_P256;
  dc->path = NULL;
  dc->pwd_hash = NULL;
  dc->optional_auth_user = NULL;
  dc->auth_domain = NULL;
  dc->session_timeout = AUTH_MUTUAL_SESSION_TIMEOUT;
  dc->validation_host = NULL;
  
  return dc;
}

static const char *add_authn_provider(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;
  authn_provider_list *newp;

  newp = apr_pcalloc(cmd->pool, sizeof(authn_provider_list));
  newp->provider_name = apr_pstrdup(cmd->pool, arg);

  newp->provider = ap_lookup_provider(AUTHN_PROVIDER_GROUP, newp->provider_name, "0");

    if (newp->provider == NULL) {
        /* by the time they use it, the provider should be loaded and
           registered with us. */
        return apr_psprintf(cmd->pool,
                            "Unknown Authn provider: %s",
                            newp->provider_name);
    }

    if (!newp->provider->get_realm_hash) {
        /* if it doesn't provide the appropriate function, reject it */
        return apr_psprintf(cmd->pool,
                            "The '%s' Authn provider doesn't support "
                            "Mutual Authentication", newp->provider_name);
    }

    /* Add it to the list now. */
    if (!dc->providers) {
        dc->providers = newp;
    }
    else {
        authn_provider_list *last = dc->providers;

        while (last->next) {
            last = last->next;
        }
        last->next = newp;
    }

    return NULL;
}

static const char *set_validation_host(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;
  dc->validation_host = apr_pstrdup(cmd->pool, arg);
  return NULL;
}

static const char *set_host(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;

  ap_log_error(APLOG_MARK, APLOG_WARNING, 0,
               cmd->server, "Mutual: WARNING: AuthMutualHost is obsolete.  Use AuthMutualValidationHost.", arg);
  return set_validation_host(cmd, config, arg);
}

static const char *set_algorithm(cmd_parms *cmd, void *config, const char *arg)
{
  enum auth_mutual_algorithm algorithm;
  d_cfg *dc = (d_cfg *)config;

  algorithm = auth_mutual_algorithm_value(arg);
  if (algorithm == AUTH_MUTUAL_ALGORITHM_ERROR) {
    return "AuthMutualAlgorithm: unknown algorithm";
  }
  dc->algorithm = algorithm;
  return NULL;
}

static const char *set_path(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;

  dc->path = apr_pstrdup(cmd->pool, arg);
  return NULL;
}

static const char *set_pwd_hash(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;

  if (strcmp(arg, "none") == 0) {
    dc->pwd_hash = NULL;
  } else if (strcmp(arg, "md5") == 0) {
    dc->pwd_hash = "md5";
  } else if (strcmp(arg, "digest-md5") == 0) {
    dc->pwd_hash = "digest-md5";
  } else if (strcmp(arg, "sha1") == 0) {
    dc->pwd_hash = "sha1";
  } else {
    return apr_psprintf(cmd->pool, "invalid AuthMutualPwdHash algorithm: %s", arg);
  }
  return NULL;
}

static const char *set_optional_auth(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;

  if (strcmp(arg, "none") == 0) {
    dc->optional_auth_user = NULL;
  } else {
    dc->optional_auth_user = apr_pstrdup(cmd->pool, arg);
  }
  return NULL;
}

static const char *set_auth_domain(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;

  if (!arg || strcmp(arg, "") == 0) {
    dc->auth_domain = NULL;
  } else {
    dc->auth_domain = apr_pstrdup(cmd->pool, arg);
  }
  return NULL;
}

static const char *set_session_timeout(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;
  int v = AUTH_MUTUAL_SESSION_TIMEOUT;

  if (arg) {
    v = atoi(arg);
  }
  if (v < 60 || v >= 7*24*60*60) {
    ap_log_error(APLOG_MARK, APLOG_WARNING, 0,
                 cmd->server, "Mutual: WARNING: specified session timeout %d sec is out of range.", v);
    v = AUTH_MUTUAL_SESSION_TIMEOUT;
  }
  dc->session_timeout = v;
  return NULL;
}

static const char *set_login_url(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;

  if (!arg || strcmp(arg, "") == 0) {
    dc->login_url = NULL;
  } else {
    dc->login_url = apr_pstrdup(cmd->pool, arg);
  }
  return NULL;
}

static const char *set_logout_url(cmd_parms *cmd, void *config, const char *arg)
{
  d_cfg *dc = (d_cfg *)config;

  if (!arg || strcmp(arg, "") == 0) {
    dc->logout_url = NULL;
  } else {
    dc->logout_url = apr_pstrdup(cmd->pool, arg);
  }
  return NULL;
}


static const command_rec auth_mutual_cmds[] =
{
  AP_INIT_ITERATE("AuthMutualProvider", add_authn_provider, NULL, OR_AUTHCFG,
                  "specify the auth providers for a directory or location"),
  AP_INIT_TAKE1("AuthMutualHost", set_host, NULL, OR_AUTHCFG,
                "specify the FQDN for a directory or location"),
  AP_INIT_TAKE1("AuthMutualValidationHost", set_validation_host, NULL, OR_AUTHCFG,
                "specify the URL used by validation."),
  AP_INIT_TAKE1("AuthMutualAlgorithm", set_algorithm, NULL, OR_AUTHCFG,
                "specify algorithm used"),
  AP_INIT_TAKE1("AuthMutualPath", set_path, NULL, OR_AUTHCFG,
                "specify the absolute root URL of this realm"),  
  AP_INIT_TAKE1("AuthMutualPwdHash", set_pwd_hash, NULL, OR_AUTHCFG,
                "specify the password hash algorithm [none/md5/digest-md5/sha1]"),
  AP_INIT_TAKE1("AuthMutualOptionalAuth", set_optional_auth, NULL, OR_AUTHCFG,
                "specify an anonymous user name used by optional authentication.  set 'none' to disable it."),
  AP_INIT_TAKE1("AuthMutualAuthDomain", set_auth_domain, NULL, OR_AUTHCFG,
                "set auth-domain parameter."),
  AP_INIT_TAKE1("AuthMutualSessionTimeout", set_session_timeout, NULL, OR_AUTHCFG,
                "specify session timeout time in second."),
  AP_INIT_TAKE1("AuthMutualLoginURL", set_login_url, NULL, OR_AUTHCFG,
                "specify URL of login page."),
  AP_INIT_TAKE1("AuthMutualLogoutURL", set_logout_url, NULL, OR_AUTHCFG,
                "specify URL of logout page."),
  {NULL}
};

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA auth_mutual_module = {
  STANDARD20_MODULE_STUFF, 
  create_auth_mutual_dir_config,    /* create per-dir    config structures */
  NULL,                  /* merge  per-dir    config structures */
  auth_create_config,                  /* create per-server config structures */
  NULL,                  /* merge  per-server config structures */
  auth_mutual_cmds,                  /* table of config file commands       */
  auth_register_hooks  /* register hooks                      */
};
