#ifndef MOD_AUTH_MUTUAL_H
#define MOD_AUTH_MUTUAL_H

#include "apr.h"
#include "apr_thread_mutex.h"
#include "apr_thread_rwlock.h"
#include "apr_tables.h"
#include "apr_time.h"
#include "apr_ldap.h"

#if APR_HAS_SHARED_MEMORY
#include "apr_rmm.h"
#include "apr_shm.h"
#endif


#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "apr_optional.h"
#include "ap_provider.h"
#include "mod_auth.h"

/*
 * Module local state.
 *
 * TODO: runtime data and configuration are in same struct.
 *       split configuration record away.
 */
typedef struct auth_mutual_state_t {
  apr_global_mutex_t *mutex; /* lock for shared memory */
  char *mutex_name;          /* lock name if required. */
  apr_size_t cache_bytes;    /* cache shm size */
  char *cache_file;          /* cache shm name if requred. */
#if APR_HAS_SHARED_MEMORY
  apr_shm_t *cache_shm;
  apr_rmm_t *cache_rmm;
#endif
  void *table;
} auth_mutual_state_t;

#define AUTH_MUTUAL_NC_MAX (0xFFFF)
#define AUTH_MUTUAL_NC_OVERFLOW (AUTH_MUTUAL_NC_MAX + 1)
#define AUTH_MUTUAL_NC_WND 16
typedef apr_uint32_t auth_mutual_nc_t;

#define AUTH_MUTUAL_SESSION_TIMEOUT 600

/*
 * nonce value nc is valid iff
 *  nc >= nc_min and
 *    let pos = nc - nc_min in
 *      (pos >= NC_WND
 *        or
 *      (~used_bitmap & (1 << pos))
 *
 * validate_and_update(nc, used_bitmap, nc_min)
 *   valid = false
 *   if nc_min <= nc and nc <= NC_MAX
 *     pos = nc - nc_min
 *     if pos >= NC_WND
 *       shift = pos - NC_WND + 1
 *       // dispose 'shift' count entry
 *       used_bitmap >>= shift  // Note: more than 32 bits shift is undefined in C
 *       nc_min += shift
 *
 *       valid = true
 *       pos = NC_WND - 1
 *     else
 *       if (~used_bitmap & (1 << pos)) != 0
 *         valid = true
 *       end
 *     end
 *     used_bitmap |= (1 << pos)
 *   end
 * end
 *
 * atomic operation uses CAS.
 *
 * atomic_validate_and_update_16_16(nc, ptr)
 *   do
 *     old32 = (used_bitmap16, nc_min16) = atomic_read32(ptr)
 *     (v, new_b16, new_m16) = validate_and_update(nc, used_bitmap16, nc_min16)
 *     return v if not v
 *     new32 = (new_b16, new_m16)
 *   while (atomic_cas32(ptr, old, new) != old)
 *   return v
 * end
 */
typedef union auth_mutual_nc_rec {
  struct {
    apr_uint16_t used_bitmap;
    apr_uint16_t nc_min;
  } t16;
  apr_uint32_t t32;
} auth_mutual_nc_rec;

void auth_mutual_init_nonce(auth_mutual_nc_rec *rec);
void auth_mutual_invalidate_nonce(auth_mutual_nc_rec *rec);
int auth_mutual_validate_and_update_nonce(auth_mutual_nc_t nc, auth_mutual_nc_rec *rec);

#define AUTH_MUTUAL_SID_LEN 20
#define AUTH_MUTUAL_SID_HEX_LEN 41

/*
 * Session record.
 */
typedef struct auth_mutual_session_state_t {
  char sid[AUTH_MUTUAL_SID_LEN];
  int ref_count;
  int timeout;
  apr_time_t last_use;
  struct auth_mutual_session_state_t *hash_next;
  struct auth_mutual_session_state_t *lru_next, *lru_prev;
  auth_mutual_nc_rec nc_rec;
  char *realm;
  size_t len;
  void *data;
} auth_mutual_session_state_t;

/* session management */
apr_status_t auth_mutual_session_init(auth_mutual_state_t *st, apr_pool_t *pool);
apr_status_t auth_mutual_session_init_child(auth_mutual_state_t *st, apr_pool_t *pool);

auth_mutual_session_state_t *auth_mutual_session_alloc(auth_mutual_state_t *st,
                                                       int timeout,
                                                       const char *realm,
                                                       size_t len,
                                                       char hexsid[AUTH_MUTUAL_SID_HEX_LEN]);
auth_mutual_session_state_t *auth_mutual_session_get(auth_mutual_state_t *st,
                                                     const char *hexsid);

void auth_mutual_session_release(auth_mutual_state_t *st, auth_mutual_session_state_t *p);

void auth_mutual_session_debug_dump(auth_mutual_state_t *st, request_rec *r);

void auth_mutual_init_crypto_thread(apr_pool_t *p, server_rec *s);

#endif
