#include <openssl/crypto.h>
#include <openssl/bn.h>
#include <openssl/evp.h>

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "mutual.h"
#include "mutual_primes.h"

#define MAX_Q_BYTES (4096 / 8)

/*
 * point of Z/pZ or EC
 *
 * allocated on BN_CTX stack
 */
typedef struct POINT {
  BIGNUM *x;
  BIGNUM *y; // EC only
  int unit;  // EC only
} POINT;

/*
 * allocated on heap
 */
typedef POINT POINT_data;

/*
 * run time context.
 */
typedef struct KAM3_ctx {
  struct KAM3_config *config;
  BN_CTX *bn_ctx;
  const auth_mutual_command *command;
  auth_mutual_result *result;
} KAM3_ctx;

/*
 * session data.
 */
struct auth_mutual_data {
  POINT_data wa, wb, z;
  char *user;
  int is_dummy;
};

/*
 * static configurations.
 */
typedef struct KAM3_config {
  const char *name;
  struct KAM3_func *functions;
  void *param; // KAM3_DL_param or KAM3_EC_param
  const EVP_MD *H_evp;
} KAM3_config;

/*
 * shared parameters.
 */
typedef struct KAM3_DL_param {
  BIGNUM *q;
  BIGNUM *r;
  BIGNUM *k;
  POINT   g;
  BIGNUM *gq_1;
  int bits;
} KAM3_DL_param;

typedef struct KAM3_EC_param {
  BIGNUM *p;
  BIGNUM *r;
  BIGNUM *a;
  BIGNUM *b;
  BIGNUM *cof_2;
  BIGNUM *cof_p;
  BIGNUM *u; /* (p - 3) / 4 */
  POINT  g;
  int bits;
} KAM3_EC_param;

static int isDL(KAM3_ctx *ctx);
static int isEC(KAM3_ctx *ctx);

static KAM3_DL_param *get_DL_param(KAM3_ctx *ctx)
{
  return (KAM3_DL_param *)ctx->config->param;
}

static KAM3_EC_param *get_EC_param(KAM3_ctx *ctx)
{
  return (KAM3_EC_param *)ctx->config->param;
}

/*
 * group manipulation
 */
typedef struct KAM3_func {
  // r = a * b (Z/pZ)
  // r = a + b (EC)
  int (*mul)(KAM3_ctx *ctx, POINT *r, POINT *a, POINT *b);
  // r = a^-1 (Z/pZ)
  // r = -a   (EC)
  //int (*inv)(KAM3_ctx *ctx, POINT *r, POINT *a); // client only
  // r = a^b   (Z/pZ)
  // r = [b] * a (EC)
  int (*exp)(KAM3_ctx *ctx, POINT *r, POINT *a, BIGNUM *b);
  // r == 0   (Z/pZ)
  // r == INF (EC)
  int (*isUnit)(KAM3_ctx *ctx, POINT *r);
  // generator
  POINT *(*G)(KAM3_ctx *ctx);
  
  size_t (*GE2OSx)(KAM3_ctx *ctx, unsigned char dst[MAX_Q_BYTES], POINT *x);
  int (*random)(KAM3_ctx *ctx, BIGNUM *r);
  int (*POINT_get)(KAM3_ctx *ctx, POINT *r);
  int (*POINT_from_str)(KAM3_ctx *ctx, POINT *r, char *s); // dec2bn
  char *(*POINT_to_str)(KAM3_ctx *ctx, POINT *r);

} KAM3_func;

static KAM3_func *get_KAM3_func(KAM3_ctx *ctx)
{
  return ctx->config->functions;
}

/*
 * DL functions
 */
static int mul_DL(KAM3_ctx *ctx, POINT *r, POINT *a, POINT *b)
{
  return BN_mod_mul(r->x, a->x, b->x, get_DL_param(ctx)->q, ctx->bn_ctx);
}

//static int inv_DL(KAM3_ctx *ctx, POINT *r, POINT *a);

static int exp_DL(KAM3_ctx *ctx, POINT *r, POINT *a, BIGNUM *b)
{
  return BN_mod_exp(r->x, a->x, b, get_DL_param(ctx)->q, ctx->bn_ctx);
}

static int isUnit_DL(KAM3_ctx *ctx, POINT *r)
{
  return BN_is_zero(r->x);
}

static POINT *generator_DL(KAM3_ctx *ctx)
{
  return &get_DL_param(ctx)->g;
}

static size_t GE2OSx_DL(KAM3_ctx *ctx, unsigned char dst[MAX_Q_BYTES], POINT *r)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  BIGNUM *z;
  int bits = get_DL_param(ctx)->bits;
  int len = (bits + 7) / 8;
  int xlen;
  size_t ret = 0;

  if (len > MAX_Q_BYTES) return 0;
  
  BN_CTX_start(bn_ctx);
  z = BN_CTX_get(bn_ctx);
  if (z == NULL) goto err;

  if (!BN_zero(z)) goto err;
  if (BN_cmp(r->x, z) < 0 || BN_cmp(r->x, get_DL_param(ctx)->q) >= 0) goto err;
  xlen = BN_num_bytes(r->x);
  if (xlen > len) goto err;
  memset(dst, 0, len - xlen);
  BN_bn2bin(r->x, dst + (len - xlen));
  ret = len;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

static int random_DL(KAM3_ctx *ctx, BIGNUM *r)
{
  return BN_rand_range(r, get_DL_param(ctx)->q);
}

static int POINT_get_DL(KAM3_ctx *ctx, POINT *r)
{
  r->x = BN_CTX_get(ctx->bn_ctx);
  r->y = NULL;
  r->unit = 0;
  return r->x != NULL;
}

/*
 * if src_len == -1, len = strlen(src)
 */
static int decode_base64(void *dst, int dst_len, const void *src, int src_len, int no_nl)
{
  BIO *b, *b64;
  int ret = -1;
  
  b64 = BIO_new(BIO_f_base64());
  if (!b64) return -1;
  b = BIO_new_mem_buf((void *)src, src_len);
  if (!b) goto err;
  if (no_nl) {
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
  }
  BIO_push(b64, b);

  ret = BIO_read(b64, dst, dst_len);
err:
  BIO_free_all(b64);
  return ret;
}

static char *encode_base64(const void *src, int src_len, int no_nl)
{
  BIO *b, *b64;
  char *ret = NULL, *t;
  int t_len;

  b64 = BIO_new(BIO_f_base64());
  if (!b64) return NULL;
  b = BIO_new(BIO_s_mem());
  if (!b) goto err;
  if (no_nl) {
    BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
  }
  BIO_push(b64, b);

  if (BIO_write(b64, src, src_len) != src_len) goto err;
  if (BIO_flush(b64) <= 0) goto err;

  t_len = BIO_get_mem_data(b, &t);
  ret = malloc(t_len + 1);
  if (ret == NULL) goto err;
  memcpy(ret, t, t_len);
  ret[t_len] = 0;
err:
  BIO_free_all(b64);
  return ret;
}

static int POINT_from_str_DL(KAM3_ctx *ctx, POINT *r, char *s)
{
  int bits = get_DL_param(ctx)->bits;
  int len = (bits + 7) / 8;
  char buf[MAX_Q_BYTES+1]; /* sizeof(buf) > len */
  int decode_len;

  decode_len = decode_base64(buf, sizeof(buf), s, -1, 1);
  if (decode_len != len) return 0;

  if (BN_bin2bn(buf, len, r->x) == NULL)
    return 0;
  return 1;
}

static char *POINT_to_str_DL(KAM3_ctx *ctx, POINT *r)
{
  int bits = get_DL_param(ctx)->bits;
  int len = (bits + 7) / 8;
  int xlen;
  char xbin[MAX_Q_BYTES];
  char *s, *t;
  long slen;
  BIO *b, *b64;
  int i;

  xlen = BN_num_bytes(r->x);
  if (xlen > len || len > MAX_Q_BYTES) return NULL;
  memset(xbin, 0, len - xlen);
  BN_bn2bin(r->x, xbin + (len - xlen));

  return encode_base64(xbin, len, 1);
}

static KAM3_func DL_func = {
  mul_DL,
  //inv_DL,
  exp_DL,
  isUnit_DL,
  generator_DL,
  GE2OSx_DL,
  random_DL,
  POINT_get_DL,
  POINT_from_str_DL,
  POINT_to_str_DL,
};

/*
 * EC functions
 */
static int copy_EC_point(POINT *dst, POINT *src)
{
  if (src->unit) {
    dst->unit = 1;
  } else {
    if (!BN_copy(dst->x, src->x)) return 0;
    if (!BN_copy(dst->y, src->y)) return 0;
    dst->unit = 0;
  }
  return 1;
}

static int add_EC_eq(KAM3_ctx *ctx, POINT *R, POINT *P)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  BIGNUM *p = get_EC_param(ctx)->p;
  BIGNUM *s, *t, *u;
  int ret = 0;

  BN_CTX_start(bn_ctx);
  if (R == P) {
    t = BN_CTX_get(bn_ctx);
    u = BN_CTX_get(bn_ctx);
  } else {
    t = R->x;
    u = R->y;
  }
  s = BN_CTX_get(bn_ctx);
  if (s == NULL) goto err;

  // t = Px * Px
  if (!BN_mod_mul(t, P->x, P->x, p, bn_ctx)) goto err;
  // t = t * 3
  if (!BN_set_word(u, 3)) goto err;
  if (!BN_mod_mul(t, t, u, p, bn_ctx)) goto err;
  // t = t + a
  if (!BN_mod_add(t, t, get_EC_param(ctx)->a, p, bn_ctx)) goto err;
  // u = Py + Py
  if (!BN_mod_add(u, P->y, P->y, p, bn_ctx)) goto err;
  // u = inv(u)
  if (!BN_mod_inverse(u, u, p, bn_ctx)) goto err;
  // s = t * u
  // s == (3Px^2+a)/2Py
  if (!BN_mod_mul(s, t, u, p, bn_ctx)) goto err;
  
  // t = s * s
  if (!BN_mod_mul(t, s, s, p, bn_ctx)) goto err;
  // t = t - Px
  if (!BN_mod_sub(t, t, P->x, p, bn_ctx)) goto err;
  // t = t - Px
  // t == s^2-Px-Px
  if (!BN_mod_sub(t, t, P->x, p, bn_ctx)) goto err;
  
  // u = Px - t
  if (!BN_mod_sub(u, P->x, t, p, bn_ctx)) goto err;
  // u = s * u
  if (!BN_mod_mul(u, s, u, p, bn_ctx)) goto err;
  // u = u - Py
  // u == s * (Px - t) - Py
  if (!BN_mod_sub(u, u, P->y, p, bn_ctx)) goto err;

  if (R == P) {
    if (!BN_copy(R->x, t)) goto err;
    if (!BN_copy(R->y, u)) goto err;
  }
  R->unit = 0;
  ret = 1;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

static int add_EC_ne(KAM3_ctx *ctx, POINT *R, POINT *P, POINT *Q)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  BIGNUM *p = get_EC_param(ctx)->p;
  BIGNUM *s, *t, *u;
  int ret = 0;

  BN_CTX_start(bn_ctx);
  if (R == P || R == Q) {
    t = BN_CTX_get(bn_ctx);
    u = BN_CTX_get(bn_ctx);
  } else {
    t = R->x;
    u = R->y;
  }
  s = BN_CTX_get(bn_ctx);
  if (s == NULL) goto err;

  // u = Py - Qy
  if (!BN_mod_sub(u, P->y, Q->y, p, bn_ctx)) goto err;
  // t = Px - Qx
  if (!BN_mod_sub(t, P->x, Q->x, p, bn_ctx)) goto err;
  // t = inv(t)
  if (!BN_mod_inverse(t, t, p, bn_ctx)) goto err;
  // s  = u * t
  // s  == (Py - Qy)/(Px - Qx)
  if (!BN_mod_mul(s, u, t, p, bn_ctx)) goto err;
  
  // t = s * s
  if (!BN_mod_mul(t, s, s, p, bn_ctx)) goto err;
  // t = t - Px
  if (!BN_mod_sub(t, t, P->x, p, bn_ctx)) goto err;
  // t = t - Qx
  // t == s^2-Px-Qx
  if (!BN_mod_sub(t, t, Q->x, p, bn_ctx)) goto err;
  
  // u = Px - t
  if (!BN_mod_sub(u, P->x, t, p, bn_ctx)) goto err;
  // u = s * u
  if (!BN_mod_mul(u, s, u, p, bn_ctx)) goto err;
  // u = u - Py
  // u == s * (Px - Rx) - Py
  if (!BN_mod_sub(u, u, P->y, p, bn_ctx)) goto err;

  if (R == P || R == Q) {
    if (!BN_copy(R->x, t)) goto err;
    if (!BN_copy(R->y, u)) goto err;
  }
  R->unit = 0;
  ret = 1;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

static int add_EC(KAM3_ctx *ctx, POINT *R, POINT *P, POINT *Q)
{
  if (P->unit) return copy_EC_point(R, Q);
  if (Q->unit) return copy_EC_point(R, P);
  if (BN_cmp(P->x, Q->x) == 0) {
    if (BN_cmp(P->y, Q->y) == 0) {
      if (BN_is_zero(P->y)) {
        // P = Q and P.y = 0
        R->unit = 1;
        return 1;
      } else {
        // P = Q and P.y != 0
        return add_EC_eq(ctx, R, P);
      }
    } else {
      // P.x = Q.x and P.y = -P.y
      R->unit = 1;
      return 1;
    }
  } else {
    // P.x != Q.x
    return add_EC_ne(ctx, R, P, Q);
  }
}

//static int neg_EC(KAM3_ctx *ctx, POINT *r, POINT *a);

static int mul_EC(KAM3_ctx *ctx, POINT *r, POINT *a, BIGNUM *b)
{
  int n = BN_num_bits(b);
  int i;
  r->unit = 1;

  for (i = n - 1; i >= 0; i--) {
    if (!add_EC(ctx, r, r, r)) return 0;
    if (BN_is_bit_set(b, i)) {
      if (!add_EC(ctx, r, r, a)) return 0;
    }
  }
  return 1;
}

static int isUnit_EC(KAM3_ctx *ctx, POINT *r)
{
  return r->unit;
}

static POINT *generator_EC(KAM3_ctx *ctx)
{
  return &get_EC_param(ctx)->g;
}

/*
 * test if P is valid group element.
 */
static int on_curve_EC(KAM3_ctx *ctx, POINT *P)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  BIGNUM *p = get_EC_param(ctx)->p;
  BIGNUM *s, *t;
  int ret = 0;

  if (P->unit) return 1;

  BN_CTX_start(bn_ctx);
  s = BN_CTX_get(bn_ctx);
  t = BN_CTX_get(bn_ctx);
  if (t == NULL) goto err;
  // s = Px^2
  if (!BN_mod_mul(s, P->x, P->x, p, bn_ctx)) goto err;
  // s = Px^2 + a
  if (!BN_mod_add(s, s, get_EC_param(ctx)->a, p, bn_ctx)) goto err;
  // s = Px^3 + a * Px 
  if (!BN_mod_mul(s, s, P->x, p, bn_ctx)) goto err;
  // s = Px^3 + a * Px +b
  if (!BN_mod_add(s, s, get_EC_param(ctx)->b, p, bn_ctx)) goto err;
  // t = Py^2
  if (!BN_mod_mul(t, P->y, P->y, p, bn_ctx)) goto err;
  if (BN_cmp(s, t) == 0) ret = 1;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

/*
 * decompress compressed representation
 *
 * t = sqrt(x^3 + ax + b)
 *
 * if (t & 1 == bit0) y = t
 * else               y = p - t
 */
static int decompress_EC(KAM3_ctx *ctx, BIGNUM *y, BIGNUM *x, int bit0)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  BIGNUM *p = get_EC_param(ctx)->p;
  BIGNUM *s, *t;  
  int ret = 0;

  BN_CTX_start(bn_ctx);
  s = BN_CTX_get(bn_ctx);
  t = BN_CTX_get(bn_ctx);
  if (t == NULL) goto err;
  // s = x^2
  if (!BN_mod_mul(s, x, x, p, bn_ctx)) goto err;
  // s = x^2 + a
  if (!BN_mod_add(s, s, get_EC_param(ctx)->a, p, bn_ctx)) goto err;
  // s = x^3 + a * x 
  if (!BN_mod_mul(s, s, x, p, bn_ctx)) goto err;
  // s = x^3 + a * x +b
  if (!BN_mod_add(s, s, get_EC_param(ctx)->b, p, bn_ctx)) goto err;

  // y = s^u
  if (!BN_mod_exp(y, s, get_EC_param(ctx)->u, p, bn_ctx)) goto err;
  
  // t = y * y
  if (!BN_mod_mul(t, y, y, p, bn_ctx)) goto err;
  // test s = y^2
  if (BN_cmp(s, t) != 0) goto err;
  // test bit0
  if (BN_is_bit_set(y, 0) != bit0) {
    if (!BN_mod_sub(y, p, y, p, bn_ctx)) goto err;
  }
  ret = 1;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

static size_t GE2OSx_EC(KAM3_ctx *ctx, unsigned char dst[MAX_Q_BYTES], POINT *r)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  int bits = get_EC_param(ctx)->bits;
  int len = (bits + 1 + 7) / 8;
  BIGNUM *v;
  int vlen;
  size_t ret = 0;

  if (len > MAX_Q_BYTES) return 0;
  if (r->unit) return 0;

  BN_CTX_start(bn_ctx);
  v = BN_CTX_get(bn_ctx);
  if (v ==  NULL) goto err;
  
  if (!BN_lshift1(v, r->x)) goto err;
  if (BN_is_bit_set(r->y, 0)) {
    if (!BN_add_word(v, 1)) goto err;
  }

  vlen = BN_num_bytes(v);
  if (vlen > len) goto err;
  memset(dst, 0, len - vlen);
  BN_bn2bin(v, dst + (len - vlen));
  ret = len;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

static int random_EC(KAM3_ctx *ctx, BIGNUM *r)
{
  return BN_rand_range(r, get_EC_param(ctx)->r);
}

static int POINT_get_EC(KAM3_ctx *ctx, POINT *r)
{
  r->x = BN_CTX_get(ctx->bn_ctx);
  r->y = BN_CTX_get(ctx->bn_ctx);
  r->unit = 0;
  return r->y != NULL;
}

static int POINT_from_str_EC(KAM3_ctx *ctx, POINT *r, char *src)
{
  int bits = get_EC_param(ctx)->bits;
  int len = ((bits + 1 + 7) / 8);
  int src_len = strlen(src);
  int bit0;
  
  int ret = 0;

  if (len * 2 != src_len) goto err;
  if (!BN_hex2bn(&r->x, src)) goto err;
  bit0 = BN_is_bit_set(r->x, 0);
  if (!BN_rshift1(r->x, r->x)) goto err;
  if (!decompress_EC(ctx, r->y, r->x, bit0)) goto err;
  r->unit = 0;
  ret = 1;
err:
  return ret;
}

static char *POINT_to_str_EC(KAM3_ctx *ctx, POINT *r)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  int bits = get_EC_param(ctx)->bits;
  int len = ((bits + 1 + 7) / 8) * 2;
  BIGNUM *v;
  char *sv = NULL, *p = NULL;
  int svlen;
  int i;

  if (r->unit) return NULL;

  BN_CTX_start(bn_ctx);
  v = BN_CTX_get(bn_ctx);
  if (v ==  NULL) goto err;
  
  if (!BN_lshift1(v, r->x)) goto err;
  if (BN_is_bit_set(r->y, 0)) {
    if (!BN_add_word(v, 1)) goto err;
  }
  sv = BN_bn2hex(v);
  if (sv == NULL) goto err;
  svlen = strlen(sv);
  p = malloc(len + 1);
  if (!p) goto err;
  memset(p, '0', len - svlen);
  for (i = 0; i < svlen; i++) {
    p[i + (len - svlen)] = tolower(sv[i]);
  }
  p[i + (len - svlen)] = 0;
err:
  if (sv) OPENSSL_free(sv);
  BN_CTX_end(bn_ctx);
  return p;
}

static KAM3_func EC_func = {
  add_EC,
  //neg_EC,
  mul_EC,
  isUnit_EC,
  generator_EC,
  GE2OSx_EC,
  random_EC,
  POINT_get_EC,
  POINT_from_str_EC,
  POINT_to_str_EC,
};

static int isDL(KAM3_ctx *ctx)
{
  return get_KAM3_func(ctx) == &DL_func;
}

static int isEC(KAM3_ctx *ctx)
{
  return get_KAM3_func(ctx) == &EC_func;
}

/*
 * KAM3 primitives
 */
static int KAM3_J(KAM3_ctx *ctx, POINT *r, unsigned char h[EVP_MAX_MD_SIZE], size_t h_len);
static int KAM3_D(KAM3_ctx *ctx, POINT *r, BIGNUM *x);
static int KAM3_C(KAM3_ctx *ctx, POINT *r, BIGNUM *x, POINT *v, POINT *y);
static int KAM3_T(KAM3_ctx *ctx, POINT *x);
static int KAM3_Va(KAM3_ctx *ctx, POINT *r, unsigned char *passwd, BIGNUM *xa, POINT *ya, POINT *yb);
static int KAM3_Vb(KAM3_ctx *ctx, POINT *r, BIGNUM *xb, POINT *ya, POINT *yb);

static KAM3_DL_param DL_parameters[2];
static KAM3_EC_param EC_parameters[3];
static KAM3_config all_config[AUTH_MUTUAL_N_ALGORITHM];

const char *auth_mutual_algorithm_name(enum auth_mutual_algorithm algorithm)
{
  if (0 <= algorithm && algorithm < AUTH_MUTUAL_N_ALGORITHM)
    return all_config[algorithm].name;
  else
    return NULL;
}

enum auth_mutual_algorithm auth_mutual_algorithm_value(const char *name)
{
  int i;

  for (i = 0; i < AUTH_MUTUAL_N_ALGORITHM; i++) {
    if (strcasecmp(all_config[i].name, name) == 0)
      return i;
  }
  return AUTH_MUTUAL_ALGORITHM_ERROR;
}


static int init_DL_config(KAM3_config *config,
                          KAM3_DL_param *p,
                          const char *name,
                          char *q_s, char *r_s, char *k_s, char *g_s, char *gq1_s,
                          int bits,
                          const EVP_MD *evp)
{
  p->q = p->r = p->k = p->gq_1 = NULL;
  p->g.x = p->g.y = NULL;
  
  // DL paramete 0
  if (!BN_dec2bn(&p->q, q_s)) goto err;
  if (!BN_dec2bn(&p->r, r_s)) goto err;
  if (!BN_dec2bn(&p->k, k_s)) goto err;
  if (!BN_dec2bn(&p->g.x, g_s)) goto err;
  if (!BN_dec2bn(&p->gq_1, gq1_s)) goto err;
  p->bits = bits;
  
  config->name = name;
  config->functions = &DL_func;
  config->param = p;
  config->H_evp = evp;
  return 1;
err:
  return 0;
}

static void free_DL_config(KAM3_config *config)
{
  KAM3_DL_param *p;

  p = config->param;
  if (p->q) BN_free(p->q);
  if (p->r) BN_free(p->r);
  if (p->k) BN_free(p->k);
  if (p->g.x) BN_free(p->g.x);
  if (p->gq_1) BN_free(p->gq_1);
  p->q = p->r = p->k = p->gq_1 = NULL;
  p->g.x = p->g.y = NULL;
}

static int init_EC_config(KAM3_config *config,
                          KAM3_EC_param *p,
                          const char *name,
                          char *p_s, char *r_s, char *a_s, char *b_s,
                          char *cof_2_s, char *cof_p_s,
                          char *gx_s, char *gy_s,
                          int bits,
                          const EVP_MD *evp)
{
  p->p = p->r = p->a = p->b = NULL;
  p->cof_2 = p->cof_p = NULL;
  p->u = NULL;
  p->g.x = p->g.y = NULL;
  
  if (!BN_dec2bn(&p->p, p_s)) goto err;
  if (!BN_dec2bn(&p->r, r_s)) goto err;
  if (!BN_dec2bn(&p->a, a_s)) goto err;
  if (!BN_hex2bn(&p->b, b_s)) goto err;

  if (!BN_dec2bn(&p->cof_2, cof_2_s)) goto err;
  if (!BN_dec2bn(&p->cof_p, cof_p_s)) goto err;

  if (!BN_dec2bn(&p->u, p_s)) goto err;
  if (!BN_add_word(p->u, 1)) goto err;
  if (BN_div_word(p->u, 4) != 0) goto err;  // u = (p + 1) / 4
  
  if (!BN_hex2bn(&p->g.x, gx_s)) goto err;
  if (!BN_hex2bn(&p->g.y, gy_s)) goto err;

  p->bits = bits;
  
  config->name = name;
  config->functions = &EC_func;
  config->param = p;
  config->H_evp = evp;
  return 1;
err:
  return 0;
}

static void free_EC_config(KAM3_config *config)
{
  KAM3_EC_param *p;

  p = config->param;
  if (p->p) BN_free(p->p);
  if (p->r) BN_free(p->r);
  if (p->a) BN_free(p->a);
  if (p->b) BN_free(p->b);

  if (p->cof_2) BN_free(p->cof_2);
  if (p->cof_p) BN_free(p->cof_p);
  
  if (p->g.x) BN_free(p->g.x);
  if (p->g.y) BN_free(p->g.x);

  p->p = p->r = p->a = p->b = NULL;
  p->cof_2 = p->cof_p = NULL;
  p->g.x = p->g.y = NULL;
}

int auth_mutual_init()
{
  int ret;
  const EVP_MD *evp_sha1 = EVP_sha1();
  const EVP_MD *evp_sha256 = EVP_sha256();
  const EVP_MD *evp_sha512 = EVP_sha512();

  ret = init_DL_config(&all_config[AUTH_MUTUAL_ISO11770_4_DL_2048], &DL_parameters[0],
                       "iso-kam3-dl-2048-sha256",
                       P2048_q, P2048_r, P2048_k, P2048_g, P2048_gq1, P2048_bits, evp_sha256);
  if (!ret) goto err;
  ret = init_DL_config(&all_config[AUTH_MUTUAL_ISO11770_4_DL_4096],  &DL_parameters[1],
                       "iso-kam3-dl-4096-sha512",
                       P4096_q, P4096_r, P4096_k, P4096_g, P4096_gq1, P4096_bits, evp_sha512);
  if (!ret) goto err;
  ret = init_EC_config(&all_config[AUTH_MUTUAL_ISO11770_4_EC_P256], &EC_parameters[1],
                       "iso-kam3-ec-p256-sha256",
                       EC_P256_p, EC_P256_r, EC_P256_a, EC_P256_b,
                       EC_P256_cof_2, EC_P256_cof_p, EC_P256_gx, EC_P256_gy, EC_P256_bits, evp_sha256);
  if (!ret) goto err;
  ret = init_EC_config(&all_config[AUTH_MUTUAL_ISO11770_4_EC_P521], &EC_parameters[2],
                       "iso-kam3-ec-p521-sha512",
                       EC_P521_p, EC_P521_r, EC_P521_a, EC_P521_b,
                       EC_P521_cof_2, EC_P521_cof_p, EC_P521_gx, EC_P521_gy, EC_P521_bits, evp_sha512);
  if (!ret) goto err;

  return 1;
err:
  auth_mutual_finish();
  return 0;
}

void auth_mutual_finish(void)
{
  free_DL_config(&all_config[AUTH_MUTUAL_ISO11770_4_DL_2048]);
  free_DL_config(&all_config[AUTH_MUTUAL_ISO11770_4_DL_4096]);
  free_EC_config(&all_config[AUTH_MUTUAL_ISO11770_4_EC_P256]);
  free_EC_config(&all_config[AUTH_MUTUAL_ISO11770_4_EC_P521]);
}
                        
static void free_point_data(POINT_data *p)
{
  if (p->x) BN_free(p->x);
  if (p->y) BN_free(p->y);
  p->x = NULL;
  p->y = NULL;
}

/*
 * POINT ==> POINT_data
 */
static int write_point(POINT_data *dst, POINT *p)
{
  dst->x = NULL;
  dst->y = NULL;

  if (p->x) {
    dst->x = BN_dup(p->x);
    if (!dst->x) goto err;
  }
  if (p->y) {
    dst->y = BN_dup(p->y);
    if (!dst->y) goto err;
  }
  dst->unit = p->unit;
  return 1;
err:
  free_point_data(dst);
  return 0;
}

/*
 * POINT <== POINT_data
 */
static int read_point(POINT *dst, POINT_data *p)
{
  if (p->x) {
    if (!BN_copy(dst->x, p->x)) goto err;
  } else {
    dst->x = NULL;
  }
  if (p->y) {
    if (!BN_copy(dst->y, p->y)) goto err;
  } else {
    dst->y = NULL;
  }
  dst->unit = p->unit;
  return 1;
err:
  return 0;
}



/*
 * serialize point.
 */
static size_t save_point(unsigned char *p, size_t *x_len, size_t *y_len, POINT_data *src)
{
  size_t sz_x = src->x ? BN_num_bytes(src->x) : 0;
  size_t sz_y = src->y ? BN_num_bytes(src->y) : 0;
  size_t ret = sz_x + sz_y;
  
  *x_len = sz_x;
  *y_len = sz_y;
  if (src->x) BN_bn2bin(src->x, p);
  p += sz_x;
  if (src->y) BN_bn2bin(src->y, p);
  return ret;
}

static size_t save_point_size(POINT_data *src)
{
  size_t sz_x = src->x ? BN_num_bytes(src->x) : 0;
  size_t sz_y = src->y ? BN_num_bytes(src->y) : 0;
  return sz_x + sz_y;
}

/*
 * deserialize point.
 *
 */
static int load_point(POINT_data *dst, size_t sz_x, size_t sz_y, unsigned char **pp)
{
  unsigned char *p = *pp;
  dst->x = NULL;
  dst->y = NULL;

  if (sz_x > 0) {
    dst->x = BN_bin2bn(p, sz_x, NULL);
    if (!dst->x) goto err;
    p += sz_x;
  }
  if (sz_y > 0) {
    dst->y = BN_bin2bn(p, sz_y, NULL);
    if (!dst->y) goto err;
    p += sz_y;
  }
  dst->unit = sz_x == 0 || sz_y == 0;
  *pp = p;
  return 1;
err:
  free_point_data(dst);
  return 0;
}

struct serialized_data {
  size_t wa_len_x, wa_len_y;
  size_t wb_len_x, wb_len_y;
  size_t z_len_x,  z_len_y;
  size_t user_len;
  size_t is_dummy;
  unsigned char data[1];
};

/*
 * deserialize session data
 *
 * byte array ==> auth_mutual_data
 */
auth_mutual_data *auth_mutual_load_data(void *src, size_t len)
{
  struct serialized_data *s = src;
  auth_mutual_data *p;
  unsigned char *d = s->data;
  char *user = NULL;

  size_t hdr_len = offsetof(struct serialized_data, data);
  size_t user_len;

  if (len < hdr_len) return NULL;

  p = malloc(sizeof(auth_mutual_data));
  if (p == NULL) return NULL;

  memset(p, 0, sizeof *p);
  if (!load_point(&p->wa, s->wa_len_x, s->wa_len_y, &d)) goto err;
  if (!load_point(&p->wb, s->wb_len_x, s->wb_len_y, &d)) goto err;
  if (!load_point(&p->z,  s->z_len_x,  s->z_len_y,  &d)) goto err;
  user = malloc(s->user_len);
  if (!user) goto err;
  memcpy(user, d, s->user_len);
  user[s->user_len - 1] = 0;
  p->user = user;
  p->is_dummy = s->is_dummy;

  return p;
err:
  free_point_data(&p->wa);
  free_point_data(&p->wb);
  free_point_data(&p->z);
  free(p);
  return NULL;
}

/*
 * serialize session data
 *
 * byte array <== auth_mutual_data
 */
size_t auth_mutual_save_data(auth_mutual_data *data, void *dst, size_t len)
{
  size_t hdr_len = offsetof(struct serialized_data, data);
  size_t ret = hdr_len;
  size_t user_len = strlen(data->user) + 1;
  
  ret += save_point_size(&data->wa);
  ret += save_point_size(&data->wb);
  ret += save_point_size(&data->z);
  ret += user_len;
  if (dst != NULL && len >= ret) {
    struct serialized_data *p = dst;
    unsigned char *d = p->data;
    d += save_point(d, &p->wa_len_x, &p->wa_len_y, &data->wa);
    d += save_point(d, &p->wb_len_x, &p->wb_len_y, &data->wb);
    d += save_point(d, &p->z_len_x,  &p->z_len_y,  &data->z);
    strcpy(d, data->user);
    p->user_len = user_len;
    p->is_dummy = data->is_dummy;
  }
  return ret;
}

/*
 * memo sesion data
 *
 * auth_mutual_data <== wa, wb, z, user
 *
 */
static struct auth_mutual_data *memo_session_data(POINT *wa, POINT *wb, POINT *z, char *user, int is_dummy)
{
  auth_mutual_data *p;

  p = malloc(sizeof(auth_mutual_data));
  if (p == NULL) return NULL;

  if (!write_point(&p->wa, wa)) goto err;
  if (!write_point(&p->wb, wb)) goto err;
  if (!write_point(&p->z,  z))  goto err;
  p->user = strdup(user);
  if (!p->user) goto err;
  p->is_dummy = is_dummy;
  
  return p;
err:
  free_point_data(&p->wa);
  free_point_data(&p->wb);
  free_point_data(&p->z);
  return NULL;
}

/*
 * remember session data
 *
 * auth_mutual_data ==> wa, wb, z
 */
static int remember_session_data(auth_mutual_data *p, POINT *wa, POINT *wb, POINT *z)
{
  if (!p) return 0;

  if (!read_point(wa, &p->wa)) goto err;
  if (!read_point(wb, &p->wb)) goto err;
  if (!read_point(z,  &p->z))  goto err;

  return 1;
err:
  return 0;
}

char *auth_mutual_data_to_string(auth_mutual_data *data)
{
  BIO *b;
  char *t, *ret;
  int t_len;

  b = BIO_new(BIO_s_mem());
  if (!b) return NULL;

  BIO_puts(b, "wa=(");
  BN_print(b, data->wa.x);
  BIO_puts(b, ",");
  if (data->wa.y) BN_print(b, data->wa.y);

  BIO_puts(b, "),wb=(");
  BN_print(b, data->wb.x);
  BIO_puts(b, ",");
  if (data->wb.y) BN_print(b, data->wb.y);
  
  BIO_puts(b, "),z=(");
  BN_print(b, data->z.x);
  BIO_puts(b, ",");
  if (data->z.y) BN_print(b, data->z.y);

  BIO_puts(b, ",user=");
  BIO_puts(b, data->user);
  BIO_puts(b, ",is_dummy=");
  BIO_puts(b, data->is_dummy ? "true" : "false");

  t_len = BIO_get_mem_data(b, &t);
  ret = malloc(t_len + 1);
  if (ret == NULL) goto err;
  memcpy(ret, t, t_len);
  ret[t_len] = 0;
err:
  BIO_free_all(b);
  return ret;
}

void auth_mutual_free_data(auth_mutual_data *data)
{
  if (data != NULL) {
    free_point_data(&data->wa);
    free_point_data(&data->wb);
    free_point_data(&data->z);
    if (data->user) free(data->user);
    free(data);
  }
}

void auth_mutual_free_result(auth_mutual_result *result)
{
  if (result) {
    if (result->wb)   free(result->wb);
    if (result->ob)   free(result->ob);
    if (result->user) free(result->user);
    result->wb   = NULL;
    result->ob   = NULL;
    result->user = NULL;
    if (result->data_out) {
      auth_mutual_free_data(result->data_out);
    }
    result->data_out = NULL;
#ifdef AUTH_MUTUAL_VERBOSE
    if (result->sb) {
      free(result->sb);
      result->sb = NULL;
    }
#endif
  }
}

/*
 * Process request-A1
 *
 * param: decimal string (Wa)
 * data_in->v: big endian BIGNUM binary
 *
 * STEP B1:
 *  wa = STR2I(param)
 *  test T(wa)
 *  sb = random([1..r-1])
 *  wb = C(sb, v, wa)
 *
 * STEP B2:
 *  z = Vb(sb, wa, wb)
 *
 * memoize(wa, wb, z)
 * result = I2STR(wb)
 */
static int process_A1(KAM3_ctx *ctx)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  KAM3_func *f = get_KAM3_func(ctx);

  BIGNUM *sb;
  POINT v, wa, wb, z;
  char *wb_str;
  int ret_c, ret_t;
  int ret = 0;
  int is_dummy = 0;

  BN_CTX_start(bn_ctx);
  sb = BN_CTX_get(bn_ctx);
  if (sb == NULL) goto err;
  
  if (!f->POINT_get(ctx, &v))  goto err;
  if (!f->POINT_get(ctx, &wa)) goto err;
  if (!f->POINT_get(ctx, &wb)) goto err;
  if (!f->POINT_get(ctx, &z))  goto err;
  
  /* read v */
  if (ctx->command->user == NULL) goto auth_fail;
  if (ctx->command->verifier == NULL) {
    /* user is not found in password database.
     * in this case, we return normal response with
     * dummy wb value.
     *
     * v = 1        if DL
     * v = infinity if EC
     */
    if (isDL(ctx)) {
      BN_one(v.x);
    } else {
      v.unit = 1;
    }
    is_dummy = 1;
  } else {
    if (!f->POINT_from_str(ctx, &v, ctx->command->verifier)) goto auth_fail;
  }

  /* read wa */
  if (ctx->command->wa == NULL) goto auth_fail;
  if (!f->POINT_from_str(ctx, &wa, ctx->command->wa)) goto auth_fail;
  
  /* test T(wa) */
  ret_t = KAM3_T(ctx, &wa);
  if (!ret_t) goto err;
  if (ret_t < 0) goto auth_fail;
  
  do {
    /* sb = random([1..r-1]); wb = C(sb, v, wa) */
    if (!f->random(ctx, sb)) goto err;
    ret_c = KAM3_C(ctx, &wb, sb, &v, &wa);
    if (!ret_c) goto err;
    /* C returns -1 if not suitable */
  } while (ret_c < 0);

  /* z = Vb(sb, wa, wb) */
  if (!KAM3_Vb(ctx, &z, sb, &wa, &wb)) goto err;
#ifdef AUTH_MUTUAL_VERBOSE
  {
    char *p = BN_bn2dec(sb);
    if (p) {
      ctx->result->sb = strdup(p);
      OPENSSL_free(p);
    }
  }
#endif
  /*
   * this function has two return value.
   *
   * ctx->result->data_out <== [wa, wb, z, user]
   * ctx->result->wb       <== wb
   */

  /* write ctx->result->data_out */
  ctx->result->data_out = memo_session_data(&wa, &wb, &z, ctx->command->user, is_dummy);
  if (!ctx->result->data_out) goto err;

  /* write ctx->result->wb */
  wb_str = f->POINT_to_str(ctx, &wb);
  ctx->result->wb = wb_str;
  if (wb_str == NULL) {
    auth_mutual_free_data(ctx->result->data_out);
    ctx->result->data_out = NULL;
    goto err;
  }
  ctx->result->result = AUTH_MUTUAL_RES_401_B1;

auth_fail:
  ret = 1;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

#if CHAR_BIT != 8
#error "CHAR_BIT != 8"
#endif

#define MAX_VI_BYTES ((sizeof(unsigned long long)*8+6)/7)

/*
 * DER encoding of unsigned integer.
 */
static size_t VI(unsigned char dst[MAX_VI_BYTES], unsigned long long v)
{
  size_t i, n;
  unsigned char buf[MAX_VI_BYTES];

  for (n = 0; n < MAX_VI_BYTES;) {
    buf[n] = v & 0x7F;
    if (n > 0) buf[n] |= 0x80;
    n++;
    v >>= 7;
    if (v == 0) break;
  }
  /* reverse */
  for (i = 0; i < n; i++) {
    dst[i] = buf[n - i - 1];
  }
  return n;
}

/*
 * returns H(hdr|GE2OSx(x[0])|...|GE2OSx(x[x_len-1]))
 */
static int key_hash_1(KAM3_ctx *ctx, unsigned char dst[EVP_MAX_MD_SIZE],
                      unsigned char hdr, POINT *x[], size_t x_len)
{
  EVP_MD_CTX md_ctx;
  unsigned char xb[MAX_Q_BYTES];
  size_t xb_len;
  size_t h_len;
  int i;

  EVP_MD_CTX_init(&md_ctx);
  EVP_DigestInit(&md_ctx, ctx->config->H_evp);
  
  EVP_DigestUpdate(&md_ctx, &hdr, 1);
  for (i = 0; i < x_len; i++) {
    xb_len = get_KAM3_func(ctx)->GE2OSx(ctx, xb, x[i]);
    EVP_DigestUpdate(&md_ctx, xb, xb_len);
  }
  
  EVP_DigestFinal(&md_ctx, dst, &h_len);
  EVP_MD_CTX_cleanup(&md_ctx);
  return h_len;
}

static void hash2hex(char dst[2 * EVP_MAX_MD_SIZE + 1], unsigned char *h, size_t h_len)
{
  size_t i;
  for (i = 0; i < h_len; i++) {
    sprintf(&dst[2 * i], "%02.2x", h[i]);
  }
}

static size_t PEM_cert_to_validation(KAM3_ctx *ctx, const char *pem_cert, char v[EVP_MAX_MD_SIZE])
{
  static const char header[] = "-----BEGIN CERTIFICATE-----\n";
  static const char trailer[] = "-----END CERTIFICATE-----\n";
  const char *p, *q;
  char *der_cert;
  int der_cert_len;
  EVP_MD_CTX md_ctx;
  size_t h_len = 0;

  if (strncmp(pem_cert, header, sizeof(header)-1) != 0) {
    return 0;
  }
  p = pem_cert + sizeof(header) - 1;
  q = strchr(p, '-');
  if (!q || strcmp(q, trailer) != 0) {
    return 0;
  }
  der_cert = malloc(q-p);
  if (!der_cert) return 0;
  der_cert_len = decode_base64(der_cert, q-p, p, q-p, 0);
  if (der_cert_len <= 0) goto err;

  EVP_MD_CTX_init(&md_ctx);
  EVP_DigestInit(&md_ctx, ctx->config->H_evp);
  EVP_DigestUpdate(&md_ctx, der_cert, der_cert_len);
  EVP_DigestFinal(&md_ctx, v, &h_len);
  EVP_MD_CTX_cleanup(&md_ctx);
err:
  if (der_cert) free(der_cert);
  return h_len;
}

/*
 * returns H(hdr|GE2OSx(x[0])|...|GE2OSx(x[x_len-1])|VI(nc)|VS(v))
 */
static int key_hash_2(KAM3_ctx *ctx, unsigned char dst[EVP_MAX_MD_SIZE],
                      unsigned char hdr, POINT *x[], size_t x_len,
                      unsigned int nc, enum validation_key_type v_type, const char *v)
{
  EVP_MD_CTX md_ctx;
  unsigned char xb[MAX_Q_BYTES];
  unsigned char vi[MAX_VI_BYTES];
  size_t xb_len;
  size_t h_len = 0;

  size_t vi_len;
  char pem_v[EVP_MAX_MD_SIZE];
  const char *v2 = NULL;
  size_t v_len;
  int i;

  if (v_type != VALIDATION_KEY_STRING && v_type != VALIDATION_KEY_PEM_CERT) {
    return 0;
  }
  
  EVP_MD_CTX_init(&md_ctx);
  EVP_DigestInit(&md_ctx, ctx->config->H_evp);

  /* write the header byte */
  EVP_DigestUpdate(&md_ctx, &hdr, 1);
  /* write x[0], x[1], ... */
  for (i = 0; i < x_len; i++) {
    xb_len = get_KAM3_func(ctx)->GE2OSx(ctx, xb, x[i]);
    EVP_DigestUpdate(&md_ctx, xb, xb_len);
  }
  /* write nc */
  vi_len = VI(vi, nc);
  EVP_DigestUpdate(&md_ctx, vi, vi_len);
  /* write v */
  if (v_type == VALIDATION_KEY_STRING) {
    v2 = v;
    v_len = strlen(v2);
  } else if (v_type == VALIDATION_KEY_PEM_CERT) {
    v_len = PEM_cert_to_validation(ctx, v, pem_v);
    if (v_len == 0) goto err;
    v2 = pem_v;
  }
  if (!v2) goto err;
  vi_len = VI(vi, v_len);
  EVP_DigestUpdate(&md_ctx, vi, vi_len);
  EVP_DigestUpdate(&md_ctx, v2,  v_len);

  EVP_DigestFinal(&md_ctx, dst, &h_len);
err:
  EVP_MD_CTX_cleanup(&md_ctx);
  return h_len;
}

/*
 * Process request-A3
 *
 * oa: hex string (Oa)
 * data_in->wa, wb, z: big endian BIGNUM binary
 *
 * STEP B3:
 *  oa' = H('4'|GE2OSx(wa)|GE2OSx(wb)|GE2OSx(z))
 *  test oa == oa'
 * STEP B4:
 *  ob  = H('3'|GE2OSx(wa)|GE2OSx(wb)|GE2OSx(z))
 *
 * result = I2HEX(ob)
 */
static int process_A3(KAM3_ctx *ctx)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  KAM3_func *f = get_KAM3_func(ctx);
  
  POINT wa, wb, z;
  POINT *to_hash[3];
  unsigned char h[EVP_MAX_MD_SIZE];
  size_t h_len;
  int ret = 0;
  
  BN_CTX_start(bn_ctx);
  if (!f->POINT_get(ctx, &wa)) goto err;
  if (!f->POINT_get(ctx, &wb)) goto err;
  if (!f->POINT_get(ctx, &z))  goto err;

  ctx->result->result = AUTH_MUTUAL_RES_401_B0_STALE;
  /* read wa, wb, z */
  if (ctx->command->data_in == NULL) goto auth_fail;
  if (!remember_session_data(ctx->command->data_in, &wa, &wb, &z)) goto err;

  /* valid sid and nc are passed */
  ctx->result->result = AUTH_MUTUAL_RES_401_B0;
  ctx->result->invalidate_session = 1;

  if (ctx->command->data_in->is_dummy) goto auth_fail;

  /* calc oa' */
  to_hash[0] = &wa;
  to_hash[1] = &wb;
  to_hash[2] = &z;
  if (ctx->command->v == NULL) goto auth_fail;
  /* H(0x04|wa|wb|z|VI(nc)|VS(v)) */
  h_len = key_hash_2(ctx, h, 0x04, to_hash, 3, ctx->command->nc, ctx->command->v_type, ctx->command->v);
  if (!h_len) goto err;
  if (ctx->command->oa == NULL) goto auth_fail;
  if (isDL(ctx)) {
    char *oa2 = encode_base64(h, h_len, 1);
    int match;
    if (!oa2) goto err;
    match = strcmp(ctx->command->oa, oa2) == 0;
    free(oa2);
    if (!match) goto auth_fail;
  } else {
    char h_hex[2 * EVP_MAX_MD_SIZE + 1];
    hash2hex(h_hex, h, h_len);
    /* compare oa with oa' */
    if (strcasecmp(ctx->command->oa, h_hex) != 0) goto auth_fail;
  }
  /* calc ob */
  h_len = key_hash_2(ctx, h, 0x03, to_hash, 3, ctx->command->nc, ctx->command->v_type, ctx->command->v);
  if (!h_len) goto err;

  if (isDL(ctx)) {
    ctx->result->ob = encode_base64(h, h_len, 1);
  } else {
    char h_hex[2 * EVP_MAX_MD_SIZE + 1];
    hash2hex(h_hex, h, h_len);
    ctx->result->ob = strdup(h_hex);
  }
  ctx->result->result = AUTH_MUTUAL_RES_200_B4;
  ctx->result->user  = strdup(ctx->command->data_in->user);

  if (ctx->result->ob == NULL || ctx->result->user == NULL) {
    free(ctx->result->ob);
    free(ctx->result->user);
    ctx->result->ob = NULL;
    ctx->result->user = NULL;
    goto err;
  }
  ctx->result->invalidate_session = 0;
auth_fail:
  ret = 1;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

char *auth_mutual_hash_passwd(enum auth_mutual_algorithm algorithm,
                              const char *hostname,
                              const char *realm,
                              const char *username,
                              const char *passwd, size_t pw_len)
{
  BN_CTX *bn_ctx;
  KAM3_func *f;
  
  KAM3_ctx ctx;
  POINT v;
  char *r = NULL;

  EVP_MD_CTX md_ctx;
  unsigned char h[EVP_MAX_MD_SIZE];
  size_t h_len;
  const char *values[] = { NULL, hostname, realm, username, passwd };
  int i;

  bn_ctx = BN_CTX_new();
  if (bn_ctx == NULL) return NULL;
  
  ctx.config = &all_config[algorithm];
  ctx.bn_ctx = bn_ctx;

  f = get_KAM3_func(&ctx);
  
  BN_CTX_start(bn_ctx);
  if (!f->POINT_get(&ctx, &v)) goto err;

  EVP_MD_CTX_init(&md_ctx);
  EVP_DigestInit(&md_ctx, ctx.config->H_evp);
  
  values[0] = ctx.config->name;
  for (i = 0; i < 5; i++) {
    unsigned char vi[MAX_VI_BYTES];
    size_t vi_len;
    size_t v_len;

    if (i == 4) {
      v_len = pw_len;
    } else {
      v_len = strlen(values[i]);
    }
    vi_len = VI(vi, v_len);
    EVP_DigestUpdate(&md_ctx, vi, vi_len);
    EVP_DigestUpdate(&md_ctx, values[i], v_len);
  }
  EVP_DigestFinal(&md_ctx, h, &h_len);

  EVP_MD_CTX_cleanup(&md_ctx);
  
  if (!KAM3_J(&ctx, &v, h, h_len)) goto err;
  r = f->POINT_to_str(&ctx, &v);

err:
  BN_CTX_end(bn_ctx);
  BN_CTX_free(bn_ctx);
  return r;
}
/*
int auth_mutual_process(enum auth_mutual_config config,
                        int cmd, char *param, char *user, char *verifier,
                        auth_mutual_data *data_in, auth_mutual_result *result)*/
int auth_mutual_process(const auth_mutual_command *cmd,
                        auth_mutual_result *result)
{
  BN_CTX *bn_ctx;
  KAM3_ctx ctx;
  int ret = 0;

  result->result   = AUTH_MUTUAL_RES_401_B0;
  result->wb       = NULL;
  result->ob       = NULL;
  result->user     = NULL;
  result->invalidate_session = 0;
  result->data_out = NULL;
#ifdef AUTH_MUTUAL_VERBOSE
  result->sb = NULL;
#endif

  bn_ctx = BN_CTX_new();
  if (bn_ctx == NULL) return 0;
  
  ctx.config = &all_config[cmd->algorithm];
  ctx.bn_ctx = bn_ctx;
  ctx.command = cmd;
  ctx.result = result;

  switch (cmd->req) {
  case AUTH_MUTUAL_REQ_A1:
    ret = process_A1(&ctx);
    break;
  case AUTH_MUTUAL_REQ_A3:
    ret = process_A3(&ctx);
    break;
  default:
    break;
  }
  BN_CTX_free(bn_ctx);
  return ret;
}

/*
 *
 * KAM3 primitives implementation
 *
 */

static int BS2I(BIGNUM *r, unsigned char *bs, unsigned int len)
{
  return BN_bin2bn(bs, len ,r)? 1: 0;
}

/*
 * returns g^{BS2I(H(passwd))} mod q
 *
 * Note: parameter h is H(passwd).
 */
static int KAM3_J(KAM3_ctx *ctx, POINT *r, unsigned char h[EVP_MAX_MD_SIZE], size_t h_len)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  KAM3_func *f = get_KAM3_func(ctx);
  
  BIGNUM *a;
  int ret = 0;

  BN_CTX_start(bn_ctx);
  a = BN_CTX_get(bn_ctx);
  if (a == NULL) goto err;

  if (!BS2I(a, h, h_len)) goto err;

  // r = g^a
  if (!f->exp(ctx, r, f->G(ctx), a)) goto err;

  ret = 1;
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

/*
 * return s g^x mod q
 */
static int KAM3_D(KAM3_ctx *ctx, POINT *r, BIGNUM *x)
{
  KAM3_func *f = get_KAM3_func(ctx);
  return f->exp(ctx, r, f->G(ctx), x);
}

/*
 * returns (v * y^e)^x mod q
 *   where e = BS2I(H('1'|GE2OSx(y)))
 */
static int KAM3_C(KAM3_ctx *ctx, POINT *r, BIGNUM *x, POINT *v, POINT *y)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  KAM3_func *f = get_KAM3_func(ctx);
  
  BIGNUM *e;
  POINT   a, b;
  EVP_MD_CTX md_ctx;
  unsigned char h[EVP_MAX_MD_SIZE];
  size_t h_len;
  int ret = 0;

  BN_CTX_start(bn_ctx);
  e = BN_CTX_get(bn_ctx);
  if (e == NULL) goto err;
  if (!f->POINT_get(ctx, &a)) goto err;
  if (!f->POINT_get(ctx, &b)) goto err;
  
  /* e = BS2I(H(...)) */
  h_len = key_hash_1(ctx, h, 1, &y, 1);
  if (!BS2I(e, h, h_len)) goto err;

  /* a = y^e   */
  if (!f->exp(ctx, &a,  y,  e)) goto err;
  /* b = v * a */
  if (!f->mul(ctx, &b,  v, &a)) goto err;
  /* r = b^x   */
  if (!f->exp(ctx,  r, &b,  x)) goto err;

  if (isDL(ctx)) {
    /* e = q-r */
    if (!BN_sub(e, get_DL_param(ctx)->q, r->x)) goto err;
    /* r == 1 || r == q-1 */
    if (BN_is_one(r->x) || BN_is_one(e)) {
      ret = -1;
    } else {
      ret = 1;
    }
  } else if (isEC(ctx)) {
    if (!BN_set_word(e, 4)) goto err;
    // a = [4]*r
    if (!f->exp(ctx, &a, r, e)) goto err;
    if (f->isUnit(ctx, &a)) {
      ret = -1;
    } else {
      ret = 1;
    }
  } else {
    goto err;
  }
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

/*
 * return 1 < x && x < q-1
 */
static int KAM3_T(KAM3_ctx *ctx, POINT *x)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  KAM3_func *f = get_KAM3_func(ctx);
  
  int ret = 0;
  
  BN_CTX_start(bn_ctx);
  if (isDL(ctx)) {
    BIGNUM *a, *one;
    a = BN_CTX_get(bn_ctx);
    one = BN_CTX_get(bn_ctx);
    if (one == NULL) goto err;

    /* one = 1 */
    if (!BN_one(one)) goto err;
    /* a = q-one */
    if (!BN_sub(a, get_DL_param(ctx)->q, one)) goto err;
    /* x <= one || x >= q-one */
    if (BN_cmp(x->x, one) <= 0 || BN_cmp(x->x, a) >= 0) {
      ret = -1;
    } else {
      ret = 1;
    }
  } else if (isEC(ctx)) {
    BIGNUM *cof = get_EC_param(ctx)->cof_2;
    POINT a;

    if (!on_curve_EC(ctx, x)) goto err;
    if (!f->POINT_get(ctx, &a)) goto err;
    // a = [cof_2]*x
    if (!f->exp(ctx, &a, x, cof)) goto err;
    if (f->isUnit(ctx, &a)) {
      ret = -1;
    } else {
      ret = 1;
    }
  } else {
    goto err;
  }
err:
  BN_CTX_end(bn_ctx);
  return ret;
}

static int KAM3_Va_DL(KAM3_ctx *ctx, POINT *r, unsigned char *passwd, BIGNUM *xa, POINT *ya, POINT *yb)
{
  // Client only.
  return 0;
}

/*
 * returns (ya * g^d)^xb mod q
 *   where d = BS2I(H('2'|GE2OSx(ya)|GE2OSx(yb)))
 */
static int KAM3_Vb(KAM3_ctx *ctx, POINT *r, BIGNUM *xb, POINT *ya, POINT *yb)
{
  BN_CTX *bn_ctx = ctx->bn_ctx;
  KAM3_func *f = get_KAM3_func(ctx);

  BIGNUM *d;
  POINT   a;
  POINT *to_hash[2];
  unsigned char h[EVP_MAX_MD_SIZE];
  size_t h_len;
  int ret = 0;

  BN_CTX_start(bn_ctx);
  d = BN_CTX_get(bn_ctx);
  if (d == NULL) goto err;
  if (!f->POINT_get(ctx, &a)) goto err;
  
  /* d = BS2I(H(...)) */
  to_hash[0] = ya;
  to_hash[1] = yb;
  h_len = key_hash_1(ctx, h, 2, to_hash, 2);
  if (!BS2I(d, h, h_len)) goto err;

  /* r = g^d  */
  if (!f->exp(ctx, r, f->G(ctx), d)) goto err;
  /* a = ya*r */
  if (!f->mul(ctx, &a, ya, r)) goto err;
  /* r = a^xb */
  if (!f->exp(ctx, r, &a, xb)) goto err;
  
  ret = 1;
err:
  BN_CTX_end(bn_ctx);
  return ret;

}

#ifdef MUTUAL_TEST
//
// test case
//
//
//
#include <assert.h>

static KAM3_config test_config;
static KAM3_EC_param test_EC_param;

static int EC_eq(POINT *x, POINT *y)
{
  if (x->unit) {
    return y->unit;
  } else if (y->unit) {
    return 0;
  } else {
    return BN_cmp(x->x, y->x) == 0 && BN_cmp(x->y, y->y) == 0;
  }
}

static void EC_set_word(POINT *p, int x, int y)
{
  BN_set_word(p->x, x);
  BN_set_word(p->y, y);
  p->unit = 0;
}

int main()
{
  BN_CTX *bn_ctx;
  KAM3_func *f;
  
  KAM3_ctx ctx;
  POINT a, b, c, d, e;
  BIGNUM *x;
  char *r = NULL;

  init_EC_config(&test_config, &test_EC_param, "11", "16", "2", "1", "1", "1", "0", "1");

  bn_ctx = BN_CTX_new();
  x = BN_CTX_get(bn_ctx);
  if (bn_ctx == NULL) return 0;
  
  ctx.config = &test_config;
  ctx.bn_ctx = bn_ctx;
  ctx.user = NULL;
  ctx.verifier = NULL;
  ctx.result = NULL;

  f = get_KAM3_func(&ctx);
  f->POINT_get(&ctx, &a);
  f->POINT_get(&ctx, &b);
  f->POINT_get(&ctx, &c);
  f->POINT_get(&ctx, &d);
  f->POINT_get(&ctx, &e);
  
  a.unit = 1;
  assert(f->isUnit(&ctx, &a));
  b.unit = 0;
  assert(!f->isUnit(&ctx, &b));
  
  EC_set_word(&b, 8, 10);
  assert(EC_eq(&a, &a));
  assert(EC_eq(&b, &b));
  assert(!EC_eq(&b, &a));

  f->mul(&ctx, &c, &a, &b);
  assert(EC_eq(&b, &c));
  f->mul(&ctx, &c, &b, &a);
  assert(EC_eq(&b, &c));

  EC_set_word(&a, 5, 9);
  EC_set_word(&b, 1, 2);
  EC_set_word(&c, 6, 3);
  EC_set_word(&d, 3, 1);

  f->mul(&ctx, &e, &a, &a);
  puts(f->POINT_to_str(&e));
  assert(EC_eq(&e, &b));

  f->mul(&ctx, &e, &a, &b);
  assert(EC_eq(&e, &c));

  f->mul(&ctx, &e, &b, &a);
  assert(EC_eq(&e, &c));

  f->mul(&ctx, &e, &c, &a);
  assert(EC_eq(&e, &d));

  f->mul(&ctx, &e, &a, &c);
  assert(EC_eq(&e, &d));

  f->mul(&ctx, &e, &b, &b);
  assert(EC_eq(&e, &d));

  BN_set_word(x, 1);
  f->exp(&ctx, &e, &a, x);
  assert(EC_eq(&e, &a));

  BN_set_word(x, 2);
  f->exp(&ctx, &e, &a, x);
  assert(EC_eq(&e, &b));

  BN_set_word(x, 3);
  f->exp(&ctx, &e, &a, x);
  assert(EC_eq(&e, &c));

  BN_set_word(x, 4);
  f->exp(&ctx, &e, &a, x);
  assert(EC_eq(&e, &d));

  BN_set_word(x, 2);
  f->exp(&ctx, &e, &b, x);
  assert(EC_eq(&e, &d));

  BN_set_word(x, 16);
  f->exp(&ctx, &e, &a, x);
  assert(f->isUnit(&ctx, &e));

  return 0;
}

#endif
