#ifndef MOD_AUTH_MUTUAL_MUTUAL_H
#define MOD_AUTH_MUTUAL_MUTUAL_H

/*
 * Stateless authentication kernel.
 */

/*
 * configuration parameter.
 */
enum auth_mutual_algorithm {
  AUTH_MUTUAL_ALGORITHM_ERROR = -1,
  AUTH_MUTUAL_ISO11770_4_EC_P256 = 0,
  AUTH_MUTUAL_ISO11770_4_EC_P521,
  AUTH_MUTUAL_ISO11770_4_DL_2048,
  AUTH_MUTUAL_ISO11770_4_DL_4096,
  AUTH_MUTUAL_N_ALGORITHM
};

const char *auth_mutual_algorithm_name(enum auth_mutual_algorithm algorithm);
enum auth_mutual_algorithm auth_mutual_algorithm_value(const char *name);

/*
 * session data.
 */
typedef struct auth_mutual_data auth_mutual_data;

enum auth_mutual_req {
  AUTH_MUTUAL_REQ_A1,
  AUTH_MUTUAL_REQ_A3
};

enum auth_mutual_res {
  AUTH_MUTUAL_RES_401_B0,
  AUTH_MUTUAL_RES_401_B0_STALE,
  AUTH_MUTUAL_RES_401_B1,
  AUTH_MUTUAL_RES_200_B4
};

enum validation_key_type {
  VALIDATION_KEY_STRING,
  VALIDATION_KEY_PEM_CERT
};

typedef struct auth_mutual_command {
  enum auth_mutual_algorithm algorithm;
  enum auth_mutual_req req;
  char *user;
  char *verifier;
  char *wa;
  char *oa;
  unsigned int nc;
  enum validation_key_type v_type;
  const char *v;
  auth_mutual_data *data_in;
} auth_mutual_command;

/*
 * response data.
 */
typedef struct auth_mutual_result {
  enum auth_mutual_res result;
  char *wb;
  char *ob;
  char *user;
  int invalidate_session;
  auth_mutual_data *data_out;
#ifdef AUTH_MUTUAL_VERBOSE
  char *sb;
#endif
} auth_mutual_result;

int auth_mutual_init(void);
void auth_mutual_finish(void);

/*
 * data serializer/deserializer
 */
auth_mutual_data *auth_mutual_load_data(void *src, size_t len);
size_t auth_mutual_save_data(auth_mutual_data *data, void *dst, size_t len);

void auth_mutual_free_data(auth_mutual_data *data);
void auth_mutual_free_result(auth_mutual_result *result);

/*
 * format session data to string.
 * used by debug logging.
 * must free string returned.
 */
char *auth_mutual_data_to_string(auth_mutual_data *data);

char *auth_mutual_hash_passwd(enum auth_mutual_algorithm algorithm,
                              const char *hostname,
                              const char *realm,
                              const char *username,
                              const char *passwd, size_t pw_len);
/*
 * Example:
 *  auth_mutual_init();
 *    ...
 *  OnRequest() {
 *    user = req.user;  # may be NULL
 *    verifier = lookup_passwd_db(user) if user; # may be NULL
 *    key = req.key; 
 *    if (key) {
 *      session = read_session(key);
 *      data_in = auth_mutual_load_data(session, session.length); # may be NULL
 *    }
 *    ret = auth_mutual_process(req.cmd, req.param, user, verifier, data_in, &result);
 *    if (!ret) error("500 Internal Server Error");
 *    if (result.data_out) key=write_session(auth_mutual_save_data(result.data_out));
 *    switch (result.result) {
 *    case AUTH_MUTUAL_RESULT_A1 or AUTH_MUTUAL_RESULT_A3:
 *      "401", "WWW-Authenticate: mutual req={A1 or A3} val={result.value} key={key}"
 *    case AUTH_MUTUAL_RESULT_OK:
 *      "200", "Authentication-Info: mutual val={result.value}"
 *    }
 *    auth_mutual_free_result(&result);
 *    auth_mutual_free_data(data_in);
 *  }
 *    ...
 *  auth_mutual_finish();
 */

void auth_mutual_porcess(const auth_mutual_command *cmd,
                         auth_mutual_result *result);

#endif
