#include "apr_file_io.h"
#include "apr_strings.h"
#include "apr_lib.h"
#include "apr_sha1.h"
#include "mutual.h"
#include "mod_auth_mutual.h"
#include <openssl/crypto.h>

#define SID_RAND_SIZE 4
#define HASH_TABLE_SIZE 1021

typedef struct {
  auth_mutual_session_state_t *table[HASH_TABLE_SIZE];
  auth_mutual_session_state_t *lru_top;
} hash_table;

#if AUTH_MUTUAL_SID_LEN != APR_SHA1_DIGESTSIZE
#error "AUTH_MUTUAL_SID_LEN != APR_SHA1_DIGESTSIZE"
#endif

#ifdef AUTH_MUTUAL_PROFILING
/* use debugger to read these values */
static apr_uint32_t auth_mutual_lock_delay_msec = 0;
static apr_uint32_t auth_mutual_lock_count;
#endif

static void lock_hashtable(auth_mutual_state_t *st)
{
#ifdef AUTH_MUTUAL_PROFILING
  apr_time_t start, end;
  apr_uint32_t diff_msec;

  start = apr_time_now();
  apr_global_mutex_lock(st->mutex);
  end = apr_time_now();
  diff_msec = apr_time_as_msec(end - start);
  apr_atomic_add32(&auth_mutual_lock_delay_msec, diff_msec);
  apr_atomic_inc32(&auth_mutual_lock_count);
#else
  apr_global_mutex_lock(st->mutex);
#endif
}

static void unlock_hashtable(auth_mutual_state_t *st)
{
  apr_global_mutex_unlock(st->mutex);
}

static int sid_equal(unsigned char x[AUTH_MUTUAL_SID_LEN],
                     unsigned char y[AUTH_MUTUAL_SID_LEN])
{
  return memcmp(x, y, AUTH_MUTUAL_SID_LEN) == 0;
}

static void generate_sid(unsigned char sid[AUTH_MUTUAL_SID_LEN])
{
  unsigned char r[SID_RAND_SIZE];
  apr_sha1_ctx_t sha1_ctx;
  
  RAND_pseudo_bytes(r, sizeof(r));
  
  apr_sha1_init(&sha1_ctx);
  apr_sha1_update_binary(&sha1_ctx, r, sizeof(r));
  apr_sha1_final(sid, &sha1_ctx);
}

static void hex_encode_sid(char hexsid[AUTH_MUTUAL_SID_HEX_LEN], unsigned char sid[AUTH_MUTUAL_SID_LEN])
{
  int i;
  
  for (i = 0; i < AUTH_MUTUAL_SID_LEN; i++) {
    int hi = (sid[i] & 0xF0) >> 4;
    int lo = (sid[i] & 0x0F);
    hexsid[i * 2]     = hi >= 10? hi - 10 + 'A': hi + '0';
    hexsid[i * 2 + 1] = lo >= 10? lo - 10 + 'A': lo + '0';
  }
  hexsid[i * 2] = '\0';
}

static int hex_to_int(char c)
{
  int v = -1;
  if ('a' <= c && c <= 'f')      v = c - 'a' + 10;
  else if ('A' <= c && c <= 'F') v = c - 'A' + 10;
  else if ('0' <= c && c <= '9') v = c - '0';
  return v;
}

static int hex_decode_sid(const char *hex, unsigned char sid[AUTH_MUTUAL_SID_LEN])
{
  int i;

  for (i = 0; i < AUTH_MUTUAL_SID_LEN; i++) {
    int hi, lo;

    hi = hex_to_int(hex[2 * i]);
    if (hi < 0) return 0;

    lo = hex_to_int(hex[2 * i + 1]);
    if (lo < 0) return 0;

    sid[i] = (hi << 4) + lo;
  }
  return hex[2 * i] == '\0';
}

static unsigned int hash_sid(unsigned char sid[AUTH_MUTUAL_SID_LEN])
{
  int i;
  unsigned int v = 0;

  for (i = 0; i < AUTH_MUTUAL_SID_LEN; i++) {
    v <<= 1;
    v += sid[i];
  }
  return v % HASH_TABLE_SIZE;
}

static void hash_del(auth_mutual_state_t *st, auth_mutual_session_state_t *p)
{
  hash_table *tbl = st->table;
  int hash = hash_sid(p->sid);
  auth_mutual_session_state_t *r, **q;

  r = tbl->table[hash];
  q = &tbl->table[hash];
  while (r) {
    if (r == p) {
      *q = r->hash_next;
      memset(p, 'D', sizeof(auth_mutual_session_state_t) + p->len);
      apr_rmm_free(st->cache_rmm, apr_rmm_offset_get(st->cache_rmm, p));
      break;
    } else {
      q = &r->hash_next;
      r = r->hash_next;
    }
  }
}

static void hash_gc(auth_mutual_state_t *st)
{
  size_t arena_sz = st->cache_bytes - sizeof(hash_table);
  hash_table *tbl = st->table;

  lock_hashtable(st);
  {
    auth_mutual_session_state_t *p, **q;
    size_t sum = 0;
    p = tbl->lru_top;
    q = &tbl->lru_top;

    while (p && sum < arena_sz / 2) {
      sum += p->len;
      q = &p->lru_next;
      p = p->lru_next;
    }

    while (p) {
      if (apr_atomic_read32(&p->ref_count) == 0) {
        *q = p->lru_next;
        if (p->lru_next) p->lru_next->lru_prev = p->lru_prev;
        hash_del(st, p);
        p = *q;
      } else {
        q = &p->lru_next;
        p = p->lru_next;
      }
    }
  }
  unlock_hashtable(st);
}

static void *shm_alloc(auth_mutual_state_t *st, apr_size_t sz)
{
  if (sz == 0)
    return NULL;
  else {
    int i;
    void *p = NULL;
    
    for (i = 0; i < 2 && p == NULL; i++) {
      apr_rmm_off_t off;
      
      lock_hashtable(st);
      off = apr_rmm_calloc(st->cache_rmm, sz);
      unlock_hashtable(st);
      
      if (off) {
        p = apr_rmm_addr_get(st->cache_rmm, off);
        break;
      }
      hash_gc(st);
    }
    return p;
  }
}

auth_mutual_session_state_t *auth_mutual_session_alloc(auth_mutual_state_t *st,
                                                       int timeout,
                                                       const char *realm,
                                                       size_t len,
                                                       char hexsid[AUTH_MUTUAL_SID_HEX_LEN])
{
  hash_table *tbl = st->table;
  size_t realm_len = strlen(realm) +  1;
  auth_mutual_session_state_t *p = shm_alloc(st, sizeof(auth_mutual_session_state_t) + len + realm_len);
  if (p == NULL) return NULL;
  p->ref_count = 1;
  p->timeout = timeout;
  p->last_use = apr_time_now();
  p->nc_rec.t32 = 0;
  p->len = len;
  p->data = (void *)(p + 1);
  p->realm = (char *)p->data + len;
  strcpy(p->realm, realm);

  for (;;) {
    int hash;
    
    generate_sid(p->sid);
    hash = hash_sid(p->sid);

    lock_hashtable(st);
    {
      auth_mutual_session_state_t *q;

      q = tbl->table[hash];
      while (q) {
        if (sid_equal(p->sid, q->sid)) {
          break;
        }
        q = q->hash_next;
      }
      if (!q) {
        p->hash_next = tbl->table[hash];
        tbl->table[hash] = p;
        p->lru_next = tbl->lru_top;
        p->lru_prev = NULL;
        if (p->lru_next) p->lru_next->lru_prev = p;
        tbl->lru_top = p;
        unlock_hashtable(st);
        break;
      }
    }
    unlock_hashtable(st);
  }
  hex_encode_sid(hexsid, p->sid);

  return p;
}

auth_mutual_session_state_t *auth_mutual_session_get(auth_mutual_state_t *st, const char *hexsid)
{
  hash_table *tbl = st->table;
  unsigned char sid[AUTH_MUTUAL_SID_LEN];
  int hash;
  auth_mutual_session_state_t *p;

  hex_decode_sid(hexsid, sid);
  hash = hash_sid(sid);
  
  lock_hashtable(st);
  {
    p = tbl->table[hash];
    while (p) {
      if (sid_equal(p->sid, sid)) break;
      p = p->hash_next;
    }
    if (p) {
      apr_time_t cur = apr_time_now();
      if (cur > p->last_use + (apr_time_t)p->timeout * 1000 * 1000) {
        p = NULL;
      } else {
        p->last_use = cur;
      }
    }
    if (p) {
      apr_atomic_inc32(&p->ref_count);
      if (p->lru_prev) {
        p->lru_prev->lru_next = p->lru_next;
        if (p->lru_next) p->lru_next->lru_prev = p->lru_prev;
        
        p->lru_prev = NULL;
        p->lru_next = tbl->lru_top;
        if (p->lru_next) p->lru_next->lru_prev = p;
        tbl->lru_top = p;
      }
    }
  }
  unlock_hashtable(st);
  return p;
}

void auth_mutual_session_release(auth_mutual_state_t *st, auth_mutual_session_state_t *p)
{
  apr_atomic_dec32(&p->ref_count);
}

apr_status_t auth_mutual_session_init(auth_mutual_state_t *st, apr_pool_t *pool)
{
  apr_status_t sts;
  hash_table *tbl;
  
  sts = apr_shm_create(&st->cache_shm, st->cache_bytes, st->cache_file, pool);
  if (sts != APR_SUCCESS) {
    return sts;
  }

  sts = apr_global_mutex_create(&st->mutex, st->mutex_name, APR_LOCK_DEFAULT, pool);
  if (sts != APR_SUCCESS) {
    return sts;
  }
    
  sts = apr_rmm_init(&st->cache_rmm, NULL, apr_shm_baseaddr_get(st->cache_shm), apr_shm_size_get(st->cache_shm), pool);
  if (sts != APR_SUCCESS) {
    return sts;
  }

  tbl = shm_alloc(st, sizeof (hash_table));
  if (!tbl) {
    return APR_ENOMEM;
  }
  st->table = tbl;
  
  return sts;
}

apr_status_t auth_mutual_session_init_child(auth_mutual_state_t *st, apr_pool_t *pool)
{
  apr_status_t sts;

  sts = apr_global_mutex_child_init(&st->mutex, st->mutex_name, pool);
  if (sts != APR_SUCCESS) {
    return;
  }
  apr_global_mutex_lock(st->mutex);
  sts = apr_rmm_attach(&st->cache_rmm, NULL, apr_shm_baseaddr_get(st->cache_shm), pool);
  apr_global_mutex_unlock(st->mutex);
  return sts;
}

void auth_mutual_session_debug_dump(auth_mutual_state_t *st, request_rec *r)
{
  char *error_fname = ap_server_root_relative(r->pool, r->server->error_fname);
  if (error_fname) {
    int i;
    for (i = 0; i < 32; i++) {
      char *dump_fname = apr_psprintf(r->pool, "%s.session_dump.%d", error_fname, i);
      apr_file_t *f;

      if (apr_file_open(&f, dump_fname, APR_WRITE|APR_CREATE|APR_BINARY|APR_EXCL,
                        APR_OS_DEFAULT, r->pool) == APR_SUCCESS) {
        apr_file_write_full(f, apr_shm_baseaddr_get(st->cache_shm), st->cache_bytes, NULL);
        apr_file_close(f);
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "auth_mutual: session memory dump: %s.", dump_fname);
        break;
      }
    }
  }
}
